import numpy as np
import pytest
from bilby.core.prior import Cosine, PriorDict, Uniform
from lightcurvelynx.math_nodes.bilby_priors import BilbyPriorNode


def test_bilby_priors_simple():
    """Test that we can generate numbers from a simple Bilby prior."""
    priors = dict(
        a=Uniform(0, 5, "a"),
        b=Uniform(0, 10, "b"),
        c=Cosine(-2, 2, "c"),
    )
    prior_dict = PriorDict(priors)
    node = BilbyPriorNode(prior=prior_dict, seed=100, node_label="sampler")
    assert "a" in node.outputs
    assert "b" in node.outputs
    assert "c" in node.outputs

    params = node.sample_parameters(num_samples=10_000)

    # "a" should be 10,000 samples from a uniform distribution between 0 and 5.
    assert params["sampler"]["a"].shape == (10_000,)
    assert np.all(params["sampler"]["a"] >= 0)
    assert np.all(params["sampler"]["a"] <= 5)
    for bin in np.linspace(0, 4, 5):
        in_bin = np.sum((params["sampler"]["a"] >= bin) & (params["sampler"]["a"] < bin + 1))
        assert 1_500 < in_bin < 2_500

    # "b" should be 10,000 samples from a uniform distribution between 0 and 10.
    assert params["sampler"]["b"].shape == (10_000,)
    assert np.all(params["sampler"]["b"] >= 0)
    assert np.all(params["sampler"]["b"] <= 10)
    for bin in np.linspace(0, 9, 10):
        in_bin = np.sum((params["sampler"]["b"] >= bin) & (params["sampler"]["b"] < bin + 1))
        assert 500 < in_bin < 1_500

    # "c" should be 10,000 samples from a cosine distribution between -2 and 2.
    assert params["sampler"]["c"].shape == (10_000,)
    assert np.all(params["sampler"]["c"] >= -2)
    assert np.all(params["sampler"]["c"] <= 2)


def test_bilby_priors_single():
    """Test that we can generate numbers from a simple Bilby prior with a single sample."""
    priors = dict(
        a=Uniform(0, 5, "a"),
        b=Uniform(0, 10, "b"),
        c=Cosine(-2, 2, "c"),
    )
    # Pass the dictionary in directly to the BilbyPriorNode.
    node = BilbyPriorNode(prior=priors, seed=100, node_label="sampler")

    params = node.sample_parameters(num_samples=1)

    # "a" should be a single sample from a uniform distribution between 0 and 5.
    assert np.isscalar(params["sampler"]["a"])
    assert params["sampler"]["a"] >= 0
    assert params["sampler"]["a"] <= 5

    # "b" should be a single sample from a uniform distribution between 0 and 10.
    assert np.isscalar(params["sampler"]["b"])
    assert params["sampler"]["b"] >= 0
    assert params["sampler"]["b"] <= 10

    # "c" should be a single sample from a cosine distribution between -2 and 2.
    assert np.isscalar(params["sampler"]["c"])
    assert params["sampler"]["c"] >= -2
    assert params["sampler"]["c"] <= 2


def test_bilby_priors_errors():
    """Test that BilbyPriorNode raises errors when given an empty prior model."""
    with pytest.raises(ValueError):
        BilbyPriorNode(prior=dict(), seed=100, node_label="sampler")



import numpy as np
import pytest
from lightcurvelynx.utils.extrapolate import (
    ConstantPadding,
    ExponentialDecay,
    LastValue,
    LinearDecay,
    ZeroPadding,
)


def test_flux_extrapolation_model():
    """Test the base class for the extrapolation methods."""
    # Create an instance of the base class
    extrapolator = ZeroPadding()

    # Test that extrapolating along wavelength returns a zero matrix
    last_wave = 1000.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_waves = np.array([1000.0, 1025.0, 1050.0, 1200.0])
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, query_waves)
    expected_result = np.zeros((3, 4))
    np.testing.assert_allclose(result, expected_result)

    # Test that we get a correctly shaped 2-D array back even if we have
    # only one query point.
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, np.array([1025.0]))
    assert result.shape == (3, 1)

    # Test that extrapolating along time returns a zero matrix
    last_time = 0.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_times = np.array([0.0, 1.0, 2.0, 5.0])
    result = extrapolator.extrapolate_time(last_time, last_flux, query_times)
    expected_result = np.zeros((4, 3))
    np.testing.assert_allclose(result, expected_result)

    # Test that we get a correctly shaped 2-D array back even if we have
    # only one query point.
    result = extrapolator.extrapolate_time(last_time, last_flux, np.array([1.0]))
    assert result.shape == (1, 3)


def test_constant_extrapolation():
    """Test that the constant extrapolation function works."""
    # Create an instance of the ConstantPadding class
    extrapolator = ConstantPadding(value=100.0)

    # Test extrapolation past the last valid point.
    last_wave = 1000.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_waves = np.array([1000.0, 1025.0, 1050.0, 1200.0])
    expected_flux = np.full((3, 4), 100.0)
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test that extrapolating along time returns a zero matrix
    last_time = 0.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_times = np.array([0.0, 1.0, 2.0, 5.0])
    result = extrapolator.extrapolate_time(last_time, last_flux, query_times)
    expected_result = np.full((4, 3), 100.0)
    np.testing.assert_allclose(result, expected_result)

    # Test that we fail if the value is not positive
    with pytest.raises(ValueError):
        _ = ConstantPadding(value=-1)


def test_last_value_extrapolation():
    """Test that the last value extrapolation function works."""
    # Create an instance of the LastValue class
    extrapolator = LastValue()

    # Test extrapolation past the last valid point.
    last_wave = 1000.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_waves = np.array([1000.0, 1025.0, 1050.0, 1200.0])
    expected_flux = np.array(
        [
            [100.0, 100.0, 100.0, 100.0],
            [200.0, 200.0, 200.0, 200.0],
            [300.0, 300.0, 300.0, 300.0],
        ]
    )
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test extrapolation along time.
    first_time = 0.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_times = np.array([-2.0, -1.0])
    expected_flux = np.array(
        [
            [100.0, 200.0, 300.0],
            [100.0, 200.0, 300.0],
        ]
    )
    result = extrapolator.extrapolate_time(first_time, last_flux, query_times)
    np.testing.assert_allclose(result, expected_flux)


def test_linear_decay_extrapolate():
    """Test that the linear decay function works."""
    decay_width = 100.0
    extrapolator = LinearDecay(decay_width=decay_width)

    # Test extrapolation past the last valid point.
    last_wave = 1000.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_waves = np.array([1000.0, 1025.0, 1050.0, 1075.0, 1100.0, 1150.0])
    expected_flux = np.array(
        [
            [100.0, 75.0, 50.0, 25.0, 0.0, 0.0],
            [200.0, 150.0, 100.0, 50.0, 0.0, 0.0],
            [300.0, 225.0, 150.0, 75.0, 0.0, 0.0],
        ]
    )
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test extrapolation before the first valid point.
    query_waves = np.array([1000.0, 975.0, 950.0, 925.0, 900.0, 850.0])
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test that we fail if the decay width is not positive
    with pytest.raises(ValueError):
        _ = LinearDecay(decay_width=-1.0)


def test_exponential_decay_extrapolate():
    """Test that the exponential decay function works."""
    extrapolator = ExponentialDecay(rate=0.1)

    # Test extrapolation past the last valid point.
    last_wave = 1000.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_waves = np.array([1000.0, 1025.0, 1050.0, 1075.0, 1100.0, 1150.0])

    t0_flux = 100.0 * np.exp([-0.0, -2.5, -5.0, -7.5, -10.0, -15.0])
    expected_flux = np.vstack((t0_flux, t0_flux * 2, t0_flux * 3))
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test extrapolation before the first valid point.
    query_waves = np.array([1000.0, 975.0, 950.0, 925.0, 900.0, 850.0])
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test that we fail if the decay rate is not positive
    with pytest.raises(ValueError):
        _ = ExponentialDecay(rate=-0.1)



import math

import jax
import numpy as np
import pytest
from lightcurvelynx.math_nodes.basic_math_node import BasicMathNode
from lightcurvelynx.math_nodes.single_value_node import SingleVariableNode


def test_basic_math_node():
    """Test that we can perform computations via a BasicMathNode."""
    node_a = SingleVariableNode("a", 10.0)
    node_b = SingleVariableNode("b", -5.0)
    node = BasicMathNode("a + b", a=node_a.a, b=node_b.b, node_label="test", backend="math")
    state = node.sample_parameters()
    assert state["test"]["function_node_result"] == 5.0

    # Try with a math function.
    node_c = SingleVariableNode("c", 1000.0)
    node = BasicMathNode("a + b - log10(c)", a=10.0, b=5.0, c=node_c.c, node_label="test", backend="math")
    state = node.sample_parameters()
    assert state["test"]["function_node_result"] == 12.0

    # Try with a second math function.
    node = BasicMathNode(
        "sqrt(a) + b - log10(c)", a=16.0, b=4.0, c=node_c.c, node_label="test", backend="math"
    )
    state = node.sample_parameters()
    assert state["test"]["function_node_result"] == 5.0

    # Test that we can reproduce the power function.
    node_d = SingleVariableNode("d", 5.0)
    node = BasicMathNode("a ** b", a=node_d.d, b=2.5, node_label="test", backend="math")
    state = node.sample_parameters()
    assert state["test"]["function_node_result"] == pytest.approx(math.pow(5.0, 2.5))


def test_basic_math_node_list_functions():
    """Test that we can list the functions available for a BasicMathNode."""
    funcs = BasicMathNode.list_functions()
    assert len(funcs) > 20


def test_basic_math_node_special_cases():
    """Test that we can handle some of the special cases for a BasicMathNode."""
    node_a = SingleVariableNode("a", 180.0)
    node = BasicMathNode("sin(deg2rad(x) + pi / 2.0)", x=node_a.a, node_label="test", backend="math")
    state = node.sample_parameters()
    assert state["test"]["function_node_result"] == pytest.approx(-1.0)


def test_basic_math_node_multi_dim():
    """Test that we can perform multidimensional computations via a BasicMathNode using numpy."""
    node_a = SingleVariableNode("a", np.array([10.0, 20.0]))
    node_b = SingleVariableNode("b", np.array([-5.0, 5.0]))
    node = BasicMathNode("a + b", a=node_a.a, b=node_b.b, node_label="test", backend="numpy")
    state = node.sample_parameters()
    assert np.array_equal(state["test"]["function_node_result"], [5.0, 25.0])

    state = node.sample_parameters(num_samples=20)
    results = state["test"]["function_node_result"]
    assert results.shape == (20, 2)
    for row in range(20):
        assert np.array_equal(results[row], [5.0, 25.0])


def test_basic_math_node_fail():
    """Test that we perform the needed checks for a math node."""
    # Imports not allowed
    with pytest.raises(ValueError):
        _ = BasicMathNode("import os")

    # Ifs not allowed (won't work with JAX)
    with pytest.raises(ValueError):
        _ = BasicMathNode("x if 1.0 else 1.0", x=2.0)

    # We only allow functions on the allow list.
    with pytest.raises(ValueError):
        _ = BasicMathNode("fake_delete_everything_no_confirm('./')")
    with pytest.raises(ValueError):
        _ = BasicMathNode("median(10, 20)")

    # All variables must be defined.
    with pytest.raises(ValueError):
        _ = BasicMathNode("x + y", x=1.0)


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_basic_math_node_error():
    """Test that we augment the error with information about the expression and parameters."""
    node = BasicMathNode("y / x", x=0.0, y=1.0)
    try:
        node.sample_parameters()
    except ZeroDivisionError as err:
        assert str(err) == "Error during math operation 'y / x' with args={'x': 0.0, 'y': 1.0}"

    node = BasicMathNode("sqrt(x)", x=-10.0)
    try:
        node.sample_parameters()
    except ValueError as err:
        assert str(err) == "Error during math operation 'np.sqrt(x)' with args={'x': -10.0}"


def test_basic_math_node_numpy():
    """Test that we can perform computations via a BasicMathNode."""
    node_a = SingleVariableNode("a", 10.0)
    node_b = SingleVariableNode("b", -5.0)
    node = BasicMathNode("a + b", a=node_a.a, b=node_b.b, node_label="test", backend="numpy")
    state = node.sample_parameters()
    assert state["test"]["function_node_result"] == 5.0

    # Try with a math function.
    node_c = SingleVariableNode("c", 1000.0)
    node = BasicMathNode("a + b - log10(c)", a=10.0, b=5.0, c=node_c.c, node_label="test", backend="numpy")
    state = node.sample_parameters()
    assert state["test"]["function_node_result"] == 12.0

    # Try with a second math function.
    node = BasicMathNode(
        "sqrt(a) + b - log10(c)", a=16.0, b=4.0, c=node_c.c, node_label="test", backend="numpy"
    )
    state = node.sample_parameters()
    assert state["test"]["function_node_result"] == 5.0

    # Test that we can reproduce the power function.
    node_d = SingleVariableNode("d", 5.0)
    node = BasicMathNode("a ** b", a=node_d.d, b=2.5, node_label="test", backend="math")
    state = node.sample_parameters()
    assert state["test"]["function_node_result"] == pytest.approx(math.pow(5.0, 2.5))


def test_basic_math_node_jax():
    """Test that we can perform computations via a BasicMathNode."""
    node_a = SingleVariableNode("a", 10.0)
    node_b = SingleVariableNode("b", -5.0)
    node = BasicMathNode("a + b", a=node_a.a, b=node_b.b, node_label="test", backend="jax")
    state = node.sample_parameters()
    assert state["test"]["function_node_result"] == 5.0

    # Try with a math function.
    node_c = SingleVariableNode("c", 1000.0)
    node = BasicMathNode("a + b - log10(c)", a=10.0, b=5.0, c=node_c.c, node_label="test", backend="jax")
    state = node.sample_parameters()
    assert state["test"]["function_node_result"] == 12.0

    # Try with a second math function.
    node = BasicMathNode(
        "sqrt(a) + b - log10(c)", a=16.0, b=4.0, c=node_c.c, node_label="test", backend="jax"
    )
    state = node.sample_parameters()
    assert state["test"]["function_node_result"] == 5.0

    # Test that we can reproduce the power function.
    node_d = SingleVariableNode("d", 5.0)
    node = BasicMathNode("a ** b", a=node_d.d, b=2.5, node_label="test", backend="math")
    state = node.sample_parameters()
    assert state["test"]["function_node_result"] == pytest.approx(math.pow(5.0, 2.5))


def test_basic_math_node_autodiff_jax():
    """Test that we can do auto-differentiation with JAX."""
    node_a = SingleVariableNode("a", 16.0, node_label="a_node")
    node_b = SingleVariableNode("b", 1000.0, node_label="b_node")

    # Create a basic math function and create tghe pytree.
    node = BasicMathNode(
        "sqrt(a) + 1.0 - log10(b)", a=node_a.a, b=node_b.b, node_label="diff_test", backend="jax"
    )
    state = node.sample_parameters()
    pytree = node.build_pytree(state)

    gr_func = jax.value_and_grad(node.generate)
    values, gradients = gr_func(pytree)
    assert values == 2.0
    assert gradients["a_node"]["a"] > 0.0
    assert gradients["b_node"]["b"] < 0.0



"""Functions for creating *fake* SED basis functions for each filter.

These functions are primarily used to test and validate bandflux only models
by allowing them to compute some SED (that will be accurate when passed through
the filter curves). Users should NOT use these functions with transforms such
as redshift.
"""

import matplotlib.pyplot as plt
import numpy as np

from lightcurvelynx.astro_utils.passbands import Passband, PassbandGroup
from lightcurvelynx.consts import lsst_filter_plot_colors


class SEDBasisModel:
    """A simple class to hold SED basis model information.

    Attributes
    ----------
    wavelengths : np.ndarray
        A 1-dimensional array of wavelengths (in Angstroms).
    sed_values : dict
        A dictionary mapping filter names to their corresponding SED values
        in each wavelength (of all_waves).
    """

    def __init__(self, wavelengths, sed_values):
        self.wavelengths = wavelengths.copy()
        self.sed_values = {k: v.copy() for k, v in sed_values.items()}
        for filter, sed in sed_values.items():
            if len(sed) != len(wavelengths):
                raise ValueError(
                    f"Length of SED values for filter {filter} ({len(sed)}) does not match "
                    f"the length of wavelengths ({len(wavelengths)})."
                )

    def __len__(self):
        """Return the number of filters in the SED basis model."""
        return len(self.sed_values)

    @property
    def filters(self):
        """Return a list of all the filter names in the SED basis model."""
        return list(self.sed_values.keys())

    @classmethod
    def from_box_approximation(cls, passbands, filters=None):
        """Create box-shaped SED basis functions. For each passband this creates a box shaped SED
        that does not overlap with any other passband. The height of the SED is normalized
        such that the total flux density will be 1.0 after passing through the passband.

        Parameters
        ----------
        passbands : PassbandGroup
            The passband group to use for defining the light curve.
        filters : list, optional
            A list of filters to use for the model. If not provided, use all filters
            in the passband group.

        Returns
        -------
        sed_basis_values : SEDBasisModel
            The basis model for this set of filters.
        """
        if isinstance(passbands, Passband):
            passbands = PassbandGroup(given_passbands=[passbands])
        if filters is None:
            filters = passbands.filters

        # Mark which wavelengths are used by each passband.
        waves_per_filter = np.zeros((len(filters), len(passbands.waves)))
        for idx, filter in enumerate(filters):
            if filter not in passbands.filters:
                raise ValueError(f"Filter {filter} not found in passband group.")

            # Get all of the wavelengths that have a non-negligible transmission value
            # for this filter and find their indices in the passband group.
            is_significant = passbands[filter].normalized_system_response[:, 1] > 1e-5
            significant_waves = passbands[filter].waves[is_significant]
            indices = np.searchsorted(passbands.waves, significant_waves)

            # Mark all non-negligible wavelengths as used by this filter.
            waves_per_filter[idx, indices] = 1.0

        # Find which wavelengths are used by multiple filters.
        filter_counts = np.sum(waves_per_filter, axis=0)

        # Create the sed values for each wavelength.
        sed_basis_values = {}
        for idx, filter in enumerate(filters):
            # Get the wavelengths that are used by ONLY this filter.
            valid_waves = (waves_per_filter[idx, :] == 1) & (filter_counts == 1)
            if np.sum(valid_waves) == 0:
                raise ValueError(
                    f"Passband {filter} has no valid wavelengths where it: a) has a non-negligible "
                    "transmission value (>0.001) and b) does not overlap with another passband."
                )

            # Compute how much flux is passed through these wavelengths of this filter
            # and use this to normalize the sed values.
            filter_sed_basis = np.zeros((1, len(passbands.waves)))
            filter_sed_basis[0, valid_waves] = 1.0

            total_flux = passbands.fluxes_to_bandflux(filter_sed_basis, filter)
            if total_flux[0] <= 0:
                raise ValueError(f"Total flux for filter {filter} is {total_flux[0]}.")
            sed_basis_values[filter] = filter_sed_basis[0, :] / total_flux[0]

        sed_basis = cls(passbands.waves, sed_basis_values)
        return sed_basis

    def compute_sed(self, filter, wavelengths=None):
        """Compute the SED values for a given filter at a single time.

        Parameters
        ----------
        filter : str
            The filter for which to compute the SED.
        wavelengths : np.ndarray, optional
            A 1-dimensional array of wavelengths (in Angstroms) at which to
            compute the SED. If None, use the internal wavelengths.

        Returns
        -------
        np.ndarray
            A 1-dimensional array of SED values at each wavelength.
        """
        if wavelengths is None:
            wavelengths = self.wavelengths

        sed_waves = np.interp(
            wavelengths,  # The query wavelengths
            self.wavelengths,  # All of the passband group's wavelengths
            self.sed_values[filter],  # The SED values at each of the passband group's wavelengths
            left=0.0,  # Do not extrapolate in wavelength
            right=0.0,  # Do not extrapolate in wavelength
        )
        return sed_waves

    def compute_sed_from_bandfluxes(self, filters, bandfluxes, wavelengths=None):
        """Compute the SED values from given bandfluxes.

        Parameters
        ----------
        filters : list of str
            A length T array of filter names for each time.
        bandfluxes : np.ndarray
            A length T array of bandflux values for each filter (in nJy).
        wavelengths : np.ndarray, optional
            A length W array of wavelengths (in Angstroms) at which to
            compute the SED. If None, use the internal wavelengths.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of observer frame SED values (in nJy).
        """
        if wavelengths is None:
            wavelengths = self.wavelengths

        filters = np.asarray(filters)
        bandfluxes = np.asarray(bandfluxes)
        if filters.shape != bandfluxes.shape:
            raise ValueError("Length of filters and bandfluxes must match.")

        # Add in the SED from each filter at each time.
        flux_density = np.zeros((len(filters), len(wavelengths)), dtype=float)
        for filter in np.unique(filters):
            if filter not in self.sed_values:
                raise ValueError(f"Filter {filter} not found in SED basis model.")

            filter_mask = np.array(filters) == filter
            sed_values = self.compute_sed(filter, wavelengths)
            flux_density[filter_mask, :] = np.outer(bandfluxes[filter_mask], sed_values)

        return flux_density

    def plot(self, ax=None, figure=None):
        """Plot the basis functions for the SED.  This is a debugging
        function to help the user understand the SEDs produced by this
        model.

        Parameters
        ----------
        ax : matplotlib.pyplot.Axes or None, optional
            Axes, None by default.
        figure : matplotlib.pyplot.Figure or None
            Figure, None by default.
        """
        if ax is None:
            if figure is None:
                figure = plt.figure()
            ax = figure.add_axes([0, 0, 1, 1])

        # Plot each passband.
        for filter_name, filter_curve in self.sed_values.items():
            color = lsst_filter_plot_colors.get(filter_name, "black")
            ax.plot(self.wavelengths, filter_curve, color=color, label=filter_name)

        # Set the x and y axis labels.
        ax.set_xlabel("Wavelength (Angstroms)")
        ax.set_ylabel("SED (nJy)")
        ax.set_title("SED Basis Functions")
        ax.legend()



import numpy as np
from scipy.spatial import KDTree


def ra_dec_to_cartesian(ra, dec):
    """
    Batch convert right ascension and declination to Cartesian coordinates.

    We use this custom function over Astropy's built-in conversion for performance reasons.
    Because we know the data is in degrees and we only need Cartesian coordinates, we can skip
    object creation and units. The results are roughly 50x faster.

    Parameters
    ----------
    ra: float or numpy.ndarray
        Right ascension in degrees.
    dec: float or numpy.ndarray
        Declination in degrees.

    Returns
    -------
    x: float or numpy.ndarray
        X coordinate.
    y: float or numpy.ndarray
        Y coordinate.
    z: float or numpy.ndarray
        Z coordinate.
    """
    ra_rad = np.radians(ra)
    dec_rad = np.radians(dec)

    x = np.cos(dec_rad) * np.cos(ra_rad)
    y = np.cos(dec_rad) * np.sin(ra_rad)
    z = np.sin(dec_rad)

    return x, y, z


def dedup_coords(ra, dec, threshold=1e-5):
    """
    Remove duplicate coordinates within a specified threshold.

    Parameters
    ----------
    ra: numpy.ndarray
        Array of right ascension values in degrees.
    dec: numpy.ndarray
        Array of declination values in degrees.
    threshold: float
        Minimum separation in degrees to consider two points as distinct.

    Returns
    -------
    unique_ra: numpy.ndarray
        Array of unique right ascension values.
    unique_dec: numpy.ndarray
        Array of unique declination values.
    unique_indices: numpy.ndarray
        Indices of the unique coordinates in the original arrays.
    """
    # Create a KD-tree for efficient nearest neighbor search.
    x, y, z = ra_dec_to_cartesian(ra, dec)
    cart_coords = np.array([x, y, z]).T
    kd_tree = KDTree(cart_coords)

    # Do a range search with the same points to find all neighbors.
    adjusted_radius = 2.0 * np.sin(0.5 * np.radians(threshold))
    close_points = kd_tree.query_ball_point(cart_coords, adjusted_radius)

    # Find unique coordinates. We keep the first occurrence of each unique point.
    # Note there will always be at least one match (the point itself).
    unique_indices = []
    for idx, matches in enumerate(close_points):
        if len(matches) == 1 or idx == np.min(matches):
            unique_indices.append(idx)
    unique_indices = np.array(unique_indices)

    unique_ra = ra[unique_indices]
    unique_dec = dec[unique_indices]
    return unique_ra, unique_dec, unique_indices



import numpy as np
import pandas as pd
import pytest
from lightcurvelynx.utils.io_utils import (
    read_grid_data,
    read_lclib_data,
    read_numpy_data,
    write_numpy_data,
    write_results_as_hats,
)
from lsdb import read_hats
from nested_pandas import NestedFrame


def test_read_write_numpy_data(tmp_path):
    """Test reading and writing numpy data."""
    data = np.arange(100, dtype=float).reshape((25, 4))

    for fmt in [".npy", ".npz", ".csv", ".ecsv", ".txt"]:
        file_path = tmp_path / f"test{fmt}"
        assert not file_path.exists()

        write_numpy_data(file_path, data)
        assert file_path.exists()

        loaded_data = read_numpy_data(file_path)
        np.testing.assert_allclose(data, loaded_data)

    # Test a file that does not exist.
    with pytest.raises(FileNotFoundError):
        _ = read_numpy_data(tmp_path / "no_such_file_here.npy")

    # Test an unsupported file format.
    with pytest.raises(ValueError):
        write_numpy_data(tmp_path / "test.invalid", data)


def test_read_write_lsdb(tmp_path):
    """Test reading and writing HATS data via LSDB."""
    outer_dict = {
        "id": [0, 1, 2],
        "ra": [10.0, 10.0001, 10.0002],
        "dec": [-10.0, -9.999, -10.0001],
        "nobs": [3, 2, 1],
        "z": [0.1, 0.2, 0.3],
    }
    inner_dict = {
        "mjd": [59000, 59001, 59002, 59000, 59001, 59000],
        "flux": [10.0, 12.0, 11.0, 15.0, 14.0, 13.0],
        "fluxerr": [1.0, 1.0, 1.0, 1.5, 1.5, 1.0],
        "filter": ["g", "r", "i", "g", "r", "i"],
    }
    nested_inds = [0, 0, 0, 1, 1, 2]
    results = NestedFrame(data=outer_dict, index=[0, 1, 2])
    nested_1 = pd.DataFrame(data=inner_dict, index=nested_inds)
    results = results.add_nested(nested_1, "lightcurve")

    # Write out the results to a temporary directory.
    out_dir = tmp_path / "lsdb_output"
    write_results_as_hats(out_dir, results, overwrite=True)
    assert out_dir.exists()

    # Check that we can read the data back in.
    loaded_results = read_hats(out_dir).compute()
    assert len(loaded_results) == len(results)
    for i in range(len(results)):
        assert results["ra"].iloc[i] == loaded_results["ra"].iloc[i]
        assert results["dec"].iloc[i] == loaded_results["dec"].iloc[i]
        pd.testing.assert_frame_equal(results["lightcurve"].iloc[i], loaded_results["lightcurve"].iloc[i])


def test_read_grid_data_good(grid_data_good_file):
    """Test that we can read a well formatted grid data file."""
    x0, x1, values = read_grid_data(grid_data_good_file, format="ascii.csv")
    x0_expected = np.array([0.0, 1.0, 2.0])
    x1_expected = np.array([1.0, 1.5])
    values_expected = np.array([[0.0, 1.0], [2.0, 3.0], [4.0, 5.0]])

    np.testing.assert_allclose(x0, x0_expected, atol=1e-5)
    np.testing.assert_allclose(x1, x1_expected, atol=1e-5)
    np.testing.assert_allclose(values, values_expected, atol=1e-5)


def test_read_grid_data_bad(grid_data_bad_file):
    """Test that we correctly handle a badly formatted grid data file."""
    # We load without a problem is validation is off.
    x0, x1, values = read_grid_data(grid_data_bad_file, format="ascii")
    assert values.shape == (3, 2)

    with pytest.raises(ValueError):
        _, _, _ = read_grid_data(grid_data_bad_file, format="ascii", validate=True)

    # We fail when loading a nonexistent file.
    with pytest.raises(FileNotFoundError):
        _, _, _ = read_grid_data("no_such_file_here", format="ascii", validate=True)


def test_read_lclib_data(test_data_dir):
    """Test reading a SNANA LCLIB data from a text file."""
    lc_file = test_data_dir / "test_lclib_data.TEXT"
    curves = read_lclib_data(lc_file)
    assert len(curves) == 3

    expected_cols = ["type", "time", "u", "g", "r", "i", "z"]
    expected_len = [20, 20, 15]
    expected_param = [
        {"TYPE": "1", "OTHER": "1"},
        {"TYPE": "1", "OTHER": "2"},
        {"TYPE": "6", "OTHER": "7"},
    ]
    for idx, curve in enumerate(curves):
        assert len(curve) == expected_len[idx]
        assert int(curve.meta["id"]) == idx
        assert curve.meta["RECUR_CLASS"] == "NON-RECUR"
        for key, value in expected_param[idx].items():
            assert curve.meta["PARVAL"][key] == value

        # We did not pick up anything in the documentation block.
        assert "PURPOSE" not in curve.meta

        # Check that the expected columns are present and type is "S" or "T"
        for col in expected_cols:
            assert col in curve.colnames
        assert np.all((curve["type"].data == "S") | (curve["type"].data == "T"))



import numpy as np
import pytest
from astropy.cosmology import Planck18
from lightcurvelynx.astro_utils.passbands import PassbandGroup
from lightcurvelynx.math_nodes.given_sampler import GivenValueList
from lightcurvelynx.models.basic_models import ConstantSEDModel
from lightcurvelynx.models.physical_model import SEDModel


def test_sed_model():
    """Test that we can create a SEDModel."""
    # Everything is specified.
    model1 = SEDModel(ra=1.0, dec=2.0, redshift=0.0, t0=1.0)
    state = model1.sample_parameters()

    assert model1.get_param(state, "ra") == 1.0
    assert model1.get_param(state, "dec") == 2.0
    assert model1.get_param(state, "redshift") == 0.0
    assert model1.get_param(state, "t0") == 1.0
    assert model1.get_param(state, "distance") is None
    assert not model1.has_valid_param("distance")
    assert model1.apply_redshift

    # Only the t0 parameter is in the PyTree.
    pytree = model1.build_pytree(state)
    assert len(pytree["SEDModel_0"]) == 1

    # We can turn off the redshift computation.
    model1.set_apply_redshift(False)
    assert not model1.apply_redshift

    # Derive the distance from the redshift. t0 is not given.
    model2 = SEDModel(ra=1.0, dec=2.0, redshift=1100.0, cosmology=Planck18)
    state = model2.sample_parameters()
    assert model2.get_param(state, "ra") == 1.0
    assert model2.get_param(state, "dec") == 2.0
    assert model2.get_param(state, "redshift") == 1100.0
    assert 13.0 * 1e12 < model2.get_param(state, "distance") < 16.0 * 1e12
    assert model2.get_param(state, "t0") is None

    # Check that the RedshiftDistFunc node has the same computed value.
    # The syntax is a bit ugly because we are checking internal state.
    model2_val = model2.get_param(state, "distance")
    func_val = model2.setters["distance"].dependency.get_param(state, "function_node_result")
    assert model2_val == func_val

    # Neither distance nor redshift are specified.
    model3 = SEDModel(ra=1.0, dec=2.0)
    state = model3.sample_parameters()
    assert model3.get_param(state, "redshift") is None
    assert model3.get_param(state, "distance") is None
    assert not model3.apply_redshift

    # Redshift is specified but cosmology is not.
    model4 = SEDModel(ra=1.0, dec=2.0, redshift=1100.0)
    state = model4.sample_parameters()
    assert model4.get_param(state, "redshift") == 1100.0
    assert model4.get_param(state, "distance") is None


def test_sed_model_multi_samples():
    """Test that we generate multiple samples from a SEDModel."""
    # Everything is specified.
    model1 = SEDModel(ra=1.0, dec=2.0, redshift=0.5, t0=1.0)
    state = model1.sample_parameters(num_samples=10)

    assert np.all(model1.get_param(state, "ra") == 1.0)
    assert np.all(model1.get_param(state, "dec") == 2.0)
    assert np.all(model1.get_param(state, "redshift") == 0.5)
    assert np.all(model1.get_param(state, "t0") == 1.0)


def test_sed_model_mask_by_time():
    """Test that we can use the default mask_by_time() function."""
    model = SEDModel(ra=1.0, dec=2.0, redshift=0.0)
    times = np.arange(-10.0, 10.0, 0.5)

    # By default use all times.
    assert np.all(model.mask_by_time(times))


def test_sed_model_evaluate_sed():
    """Test that we can evaluate a SEDModel."""
    times = np.array([0.0, 1.0, 2.0, 3.0, 4.0])
    waves = np.array([4000.0, 5000.0])
    brightness = GivenValueList([10.0, 20.0, 30.0])
    static_source = ConstantSEDModel(brightness=brightness)

    # Providing no state should give a single sample.
    flux = static_source.evaluate_sed(times, waves)
    assert flux.shape == (5, 2)
    assert np.all(flux == 10.0)

    # Doing a single sample should give a single sample.
    state = static_source.sample_parameters(num_samples=1)
    flux = static_source.evaluate_sed(times, waves, graph_state=state)
    assert flux.shape == (5, 2)
    assert np.all(flux == 20.0)

    # We can do multiple samples.
    brightness.reset()
    state = static_source.sample_parameters(num_samples=3)
    flux = static_source.evaluate_sed(times, waves, graph_state=state)
    assert flux.shape == (3, 5, 2)
    assert np.all(flux[0, :, :] == 10.0)
    assert np.all(flux[1, :, :] == 20.0)
    assert np.all(flux[2, :, :] == 30.0)


def test_sed_model_evaluate_redshift():
    """Test that if we apply redshift to a model we get different flux values."""
    times = np.array([0.0, 1.0, 2.0, 3.0, 4.0])
    waves = np.array([4000.0, 5000.0])
    static_source = ConstantSEDModel(brightness=10.0, redshift=0.5, t0=0.0)

    state = static_source.sample_parameters(num_samples=3)
    flux = static_source.evaluate_sed(times, waves, graph_state=state)
    assert flux.shape == (3, 5, 2)
    assert not np.any(flux[:, :, :] == 10.0)

    # The flux should be the same for all samples (not double applying redshift).
    assert len(np.unique(flux)) == 1


def test_sed_model_evaluate_bandflux(passbands_dir):
    """Test that band fluxes are computed correctly."""
    # It should work fine for any positive Fnu.
    f_nu = np.random.lognormal()
    static_source = ConstantSEDModel(brightness=f_nu)
    state = static_source.sample_parameters()
    passbands = PassbandGroup.from_preset(preset="LSST", table_dir=passbands_dir)
    n_passbands = len(passbands)

    times = np.arange(n_passbands, dtype=float)
    filters = np.array(sorted(passbands.passbands.keys()))

    # It should fail if no filters are provided.
    with pytest.raises(ValueError):
        _bandfluxes = static_source.evaluate_bandfluxes(passbands, times=times, filters=None, state=state)
    # It should fail if single passband is provided, but with multiple filter names.
    with pytest.raises(KeyError):
        _bandfluxes = static_source.evaluate_bandfluxes(passbands.passbands["LSST_r"], times, filters, state)

    bandfluxes = static_source.evaluate_bandfluxes(passbands, times, filters, state)
    assert bandfluxes.shape == (n_passbands,)
    np.testing.assert_allclose(bandfluxes, f_nu, rtol=1e-10)

    # If we use multiple samples, we should get a correctly sized array.
    n_samples = 21
    brightness_list = [1.5 * i for i in range(n_samples)]
    static_source2 = ConstantSEDModel(brightness=GivenValueList(brightness_list))
    state2 = static_source2.sample_parameters(num_samples=n_samples)
    bandfluxes2 = static_source2.evaluate_bandfluxes(passbands, times, filters, state2)
    assert bandfluxes2.shape == (n_samples, n_passbands)
    for idx, brightness in enumerate(brightness_list):
        np.testing.assert_allclose(bandfluxes2[idx, :], brightness, rtol=1e-10)



import numpy as np
import pytest
from citation_compass import find_in_citations
from lightcurvelynx.astro_utils.mag_flux import flux2mag
from lightcurvelynx.models.sed_template_model import (
    MultiSEDTemplateModel,
    SEDTemplate,
    SEDTemplateModel,
    SIMSEDModel,
)


def test_linear_sed_template_data() -> None:
    """Test that we can create a SEDTemplate object with linear interpolation."""
    data = np.array(
        [
            [1.0, 1000.0, 10.0],
            [1.0, 2000.0, 20.0],
            [2.0, 1000.0, 15.0],
            [2.0, 2000.0, 25.0],
            [3.0, 1000.0, 20.0],
            [3.0, 2000.0, 30.0],
        ]
    )
    data_obj = SEDTemplate(data, interpolation_type="linear", periodic=False)

    eval_times = np.array([-1.5, 1.5, 2.5, 3.5])
    eval_waves = np.array([1000.0, 2000.0])
    sed_values = data_obj.evaluate_sed(eval_times, eval_waves)
    expected_values = np.array(
        [
            [0.0, 0.0],  # 0.0 when not baseline provided
            [12.5, 22.5],
            [17.5, 27.5],
            [0.0, 0.0],  # 0.0 when not baseline provided
        ]
    )
    assert np.allclose(sed_values, expected_values)

    # We correct for sed_data_t0.
    data_obj2 = SEDTemplate(
        data,
        interpolation_type="linear",
        periodic=False,
        sed_data_t0=1.0,
    )
    sed_values = data_obj2.evaluate_sed(eval_times, eval_waves)
    expected_values = np.array(
        [
            [0.0, 0.0],  # 0.0 when not baseline provided
            [17.5, 27.5],
            [0.0, 0.0],  # 0.0 when not baseline provided
            [0.0, 0.0],  # 0.0 when not baseline provided
        ]
    )
    assert np.allclose(sed_values, expected_values)

    # Add a baseline and check that it is used outside the time range.
    baseline = np.array([1.0, 2.0])
    data_obj3 = SEDTemplate(
        data,
        interpolation_type="linear",
        periodic=False,
        baseline=baseline,
    )
    sed_values = data_obj3.evaluate_sed(eval_times, eval_waves)
    expected_values = np.array(
        [
            [1.0, 2.0],  # Baseline values when not in time range
            [12.5, 22.5],
            [17.5, 27.5],
            [1.0, 2.0],  # Baseline values when not in time range
        ]
    )
    assert np.allclose(sed_values, expected_values)

    # We fail if we use a incorrectly shaped baseline.
    with pytest.raises(ValueError):
        SEDTemplate(
            data,
            interpolation_type="linear",
            periodic=False,
            baseline=np.array([1.0, 2.0, 3.0]),
        )


def test_linear_sed_template_data_periodic() -> None:
    """Test that we can create periodic SEDTemplate object with linear interpolation."""
    data = np.array(
        [
            [1.0, 1000.0, 10.0],
            [1.0, 2000.0, 20.0],
            [2.0, 1000.0, 15.0],
            [2.0, 2000.0, 25.0],
            [3.0, 1000.0, 10.0],
            [3.0, 2000.0, 20.0],
        ]
    )
    data_obj = SEDTemplate(data, interpolation_type="linear", periodic=True)

    eval_times = np.array([0.5, 1.5, 2.25, 3.25])
    eval_waves = np.array([1000.0, 2000.0])
    sed_values = data_obj.evaluate_sed(eval_times, eval_waves)
    expected_values = np.array(
        [
            [12.5, 22.5],
            [12.5, 22.5],
            [11.25, 21.25],
            [13.75, 23.75],
        ]
    )
    assert np.allclose(sed_values, expected_values)


def test_sed_template_data_from_components() -> None:
    """Test that we can create a SEDTemplate object from separate time, wavelength, and flux arrays."""
    times = np.array([1.0, 2.0, 3.0])
    wavelengths = np.array([1000.0, 2000.0])
    fluxes = np.array(
        [
            [10.0, 20.0],
            [15.0, 25.0],
            [20.0, 30.0],
        ]
    )
    data_obj = SEDTemplate.from_components(
        times,
        wavelengths,
        fluxes,
        interpolation_type="linear",
        periodic=False,
    )

    eval_times = np.array([1.5, 2.5])
    eval_waves = np.array([1000.0, 2000.0])
    sed_values = data_obj.evaluate_sed(eval_times, eval_waves)
    expected_values = np.array(
        [
            [12.5, 22.5],
            [17.5, 27.5],
        ]
    )
    assert np.allclose(sed_values, expected_values)


def test_sed_template_data_from_file(test_data_dir):
    """Test that we can create a SEDTemplate from a file."""
    filename = test_data_dir / "truncated-salt2-h17" / "salt2_template_0.dat"
    data = SEDTemplate.from_file(filename)
    assert len(data.times) == 26
    assert len(data.wavelengths) == 401

    with pytest.raises(FileNotFoundError):
        SEDTemplate.from_file(test_data_dir / "nonexistent-file.dat")


def test_create_sed_template_model() -> None:
    """Test that we can create an SEDTemplateModel object."""
    data = np.array(
        [
            [0.0, 1000.0, 5.0],
            [0.0, 2000.0, 15.0],
            [0.0, 3000.0, 5.0],
            [1.0, 1000.0, 10.0],
            [1.0, 2000.0, 20.0],
            [1.0, 3000.0, 5.0],
            [2.0, 1000.0, 15.0],
            [2.0, 2000.0, 25.0],
            [2.0, 3000.0, 5.0],
            [3.0, 1000.0, 10.0],
            [3.0, 2000.0, 15.0],
            [3.0, 3000.0, 5.0],
        ]
    )
    model = SEDTemplateModel(data, sed_data_t0=0.0, interpolation_type="linear", periodic=False, t0=0.0)
    assert len(model.times) == 4
    assert len(model.wavelengths) == 3

    # Evaluate the model at some times and wavelengths.
    eval_times = np.array([1.5, 2.5, 3.5])
    eval_waves = np.array([1000.0, 2000.0, 2500.0])

    state = model.sample_parameters(num_samples=1)
    sed_values = model.evaluate_sed(eval_times, eval_waves, graph_state=state)
    expected_values = np.array(
        [
            [12.5, 22.5, 13.75],
            [12.5, 20.0, 12.5],
            [0.0, 0.0, 0.0],
        ]
    )
    assert np.allclose(sed_values, expected_values)

    # We fail is we provide the data in numpy array form without sed_data_t0.
    with pytest.raises(ValueError):
        _ = SEDTemplateModel(data, interpolation_type="linear", periodic=False, t0=0.0)

    # Set a non-zero t0. An evaluation time of 2.0 now corresponds to phase 0.0 in the curve.
    model2 = SEDTemplateModel(data, sed_data_t0=0.0, interpolation_type="linear", periodic=False, t0=2.0)
    state2 = model2.sample_parameters(num_samples=1)
    sed_values2 = model2.evaluate_sed(eval_times, eval_waves, graph_state=state2)
    expected_values2 = np.array(
        [
            [0.0, 0.0, 0.0],
            [7.5, 17.5, 11.25],
            [12.5, 22.5, 13.75],
        ]
    )
    assert np.allclose(sed_values2, expected_values2)


def test_create_multi_sed_template_model() -> None:
    """Test that we can create a MultiSEDTemplateModel object."""
    data1 = np.array(
        [
            [0.0, 1000.0, 5.0],
            [0.0, 2000.0, 15.0],
            [1.0, 1000.0, 10.0],
            [1.0, 2000.0, 20.0],
        ]
    )
    lc1 = SEDTemplate(data1, interpolation_type="linear", periodic=False)

    data2 = np.array(
        [
            [0.0, 1000.0, 20.0],
            [0.0, 2000.0, 30.0],
            [1.0, 1000.0, 25.0],
            [1.0, 2000.0, 35.0],
        ]
    )
    lc2 = SEDTemplate(data2, interpolation_type="linear", periodic=False)
    model = MultiSEDTemplateModel([lc1, lc2], weights=[0.25, 0.75], t0=0.0, node_label="model")
    assert len(model) == 2

    # Evaluate the model at some times and wavelengths.
    eval_times = np.array([0.5])
    eval_waves = np.array([1000.0])

    state = model.sample_parameters(num_samples=10_000)
    sed_values = model.evaluate_sed(eval_times, eval_waves, graph_state=state)
    chose_first = state["model"]["selected_template"] == 0
    assert 0.15 < np.count_nonzero(chose_first) / 10000.0 < 0.35  # Check weights are roughly correct.
    assert np.allclose(sed_values[chose_first, 0, 0], 7.5)
    assert np.allclose(sed_values[~chose_first, 0, 0], 22.5)


def test_simsed_model_compute_sed(test_data_dir) -> None:
    """Test that we can load and compute SEDs from a SIMSEDModel."""
    model = SIMSEDModel.from_dir(test_data_dir / "fake_simsed", t0=0.0, distance=10.0, node_label="model")
    assert len(model) == 2
    assert model.flux_scale == 2.0

    # Compute the SLSN SED at a specific time and wavelength.
    times = np.array([0.0, 1.0, 2.0])
    wavelengths = np.array([4500.0, 5000.0])
    sed_values = model.evaluate_sed(times, wavelengths)
    assert sed_values.shape == (3, 2)
    assert np.all(sed_values > 0.0)

    # Confirm that we have noted the data in the citations registry.
    citations = find_in_citations("SIMSED Data")
    assert len(citations) == 1
    assert str(test_data_dir / "fake_simsed") in citations[0]


def test_slsn_simsed_model(test_data_dir) -> None:
    """Test that we can compute the fluxes from a fake SLSN SIMSEDModel."""
    # Use fake data that sits around the peak of the SLSN lightcurve defined in
    # SIMSED.SLSN-I-MOSFIT/slsn0.dat.gz from https://zenodo.org/records/2612896
    # and use a similar flux_scale as given in SIMSED.SLSN-I-MOSFIT/SED.INFO to give
    # reasonable fluxes.
    data = np.array(
        [
            [0.0, 1000.0, 5.0e41],
            [0.0, 2000.0, 5.0e41],
            [1.0, 1000.0, 5.0e41],
            [1.0, 2000.0, 5.0e41],
            [2.0, 1000.0, 5.0e41],
            [3.0, 1000.0, 5.0e41],
            [2.0, 1000.0, 5.0e41],
            [3.0, 2000.0, 5.0e41],
        ]
    )
    template = SEDTemplate(data, interpolation_type="linear", periodic=False)
    model = SIMSEDModel([template], flux_scale=8.4e-41, t0=0.0, distance=10.0)
    assert len(model) == 1
    assert model.flux_scale == 8.4e-41

    # Compute the SLSN SED at a specific time and wavelength.
    times = np.array([0.5, 1.0])
    wavelengths = np.array([1200.0])
    sed_values = model.evaluate_sed(times, wavelengths)
    assert sed_values.shape == (2, 1)
    assert np.all(sed_values > 0.0)

    # We know that the expected magntiudes for a SLSN at 10 pc will be around -22.
    mag_vals = flux2mag(sed_values.flatten())
    assert np.all((mag_vals > -23.0) & (mag_vals < -21.0))



"""A collection of functions for plotting and visualization."""

import matplotlib.pyplot as plt
import numpy as np


def plot_lightcurves(
    fluxes,
    times,
    *,
    fluxerrs=None,
    filters=None,
    underlying_model=None,
    ax=None,
    figure=None,
    title=None,
    colormap=None,
    **kwargs,
):
    """Plot one or more light curves.

    Parameters
    ----------
    fluxes : numpy.ndarray
        An array of T flux values.
    times : numpy.ndarray
        A length T matrix of the times, used for setting the x axis.
        All times in MJD.
    fluxerrs : numpy.ndarray or None, optional
        A length T matrix of errors on the fluxes for error bars. If not provided
        no error bars are created. None by default.
    filters : numpy.ndarray or None, optional
        A length T matrix of filter names. If not provided all points are
        treated as coming from the same filter. None by default.
    underlying_model: dict or None, optional
        A dictionary mapping filter names to the noise free light curves for this model.
        If provided, these curves will be plotted as lines behind the data points.
        None by default.
    ax : matplotlib.pyplot.Axes or None, optional
        Axes, None by default.
    figure : matplotlib.pyplot.Figure or None
        Figure, None by default.
    title : str or None, optional
        Title of the plot. None by default.
    colormap: dict, optional
        A dictionary that provides mapping between filters and the colors to be plotted.
    **kwargs : dict
        Optional parameters to pass to the plotting function

    Returns
    -------
    ax : matplotlib.pyplot.Axes
        The axes containing the plot.
    """
    # If no axes were given create them using either the given figure or
    # a newly created one (if no figure is given).
    if ax is None:
        if figure is None:
            figure = plt.figure()
        ax = figure.add_axes([0, 0, 1, 1])

    # Set up the time array if it is not given.
    num_pts = len(fluxes)
    if len(times) != num_pts:
        raise ValueError(f"Mismatched array sizes for fluxes ({num_pts}) and times ({len(times)}).")

    # Set up a list of filters to display.
    if filters is None:
        filters = ["none"] * num_pts
        unique_filters = set(["None"])
    elif len(filters) == num_pts:
        filters = np.asarray(filters)
        unique_filters = np.unique(filters)
    else:
        raise ValueError(f"Mismatched array sizes for fluxes ({num_pts}) and filters ({len(filters)}).")

    # Check that if flux errors are given, they are the correct size.
    if fluxerrs is not None and len(fluxerrs) != num_pts:
        raise ValueError(f"Mismatched array sizes for fluxes ({num_pts}) and fluxerrs ({len(fluxerrs)}).")

    if colormap is None:
        colormap = {}
        colors = "bgrcmyk"
        for i, f in enumerate(unique_filters):
            colormap[f] = colors[i]

    # Plot the data with one line for each filter.
    for filter in unique_filters:
        filter_mask = filters == filter

        # Plot the underlying model if it is provided.
        if underlying_model is not None and filter in underlying_model:
            ax.plot(
                underlying_model["times"],
                underlying_model[filter],
                linestyle="-",
                color=colormap[filter],
                alpha=0.5,
                label=f"Model {filter}",
            )

        if fluxerrs is None:
            ax.plot(
                times[filter_mask],
                fluxes[filter_mask],
                marker="o",
                label=f"Sample {filter}",
                color=colormap[filter],
                **kwargs,
            )
        else:
            ax.errorbar(
                times[filter_mask],
                fluxes[filter_mask],
                yerr=fluxerrs[filter_mask],
                fmt="o",
                label=f"Sample {filter}",
                color=colormap[filter],
                **kwargs,
            )

    # Set the title and axis labels.
    if title is not None:
        ax.set_title(title)
    ax.set_xlabel("Time (MJD)")
    ax.set_ylabel("Flux (nJy)")

    # Only include a legend if there are at least two curves.
    if len(unique_filters) > 1:
        ax.legend()

    return ax


def plot_bandflux_lightcurves(bandflux, times=None, ax=None, figure=None, title=None):
    """Plot one or more light curves where each band is observed at each time.
    This is primarily used for visualizing non-sampled data.

    Parameters
    ----------
    bandflux : numpy.ndarray or dict
        Either a single array with the light curve or a dictionary mapping
        light curve names to the arrays of values.
    times : numpy.ndarray or None, optional
        A length T matrix of the times, used for setting the x axis. If not
        provided, uses equal spaced ticks. None by default.
    ax : matplotlib.pyplot.Axes or None, optional
        Axes, None by default.
    figure : matplotlib.pyplot.Figure or None
        Figure, None by default.
    title : str or None, optional
        Title of the plot. None by default.

    Returns
    -------
    ax : matplotlib.pyplot.Axes
        The axes containing the plot.
    """
    # If no axes were given create them using either the given figure or
    # a newly created one (if no figure is given).
    if ax is None:
        if figure is None:
            figure = plt.figure()
        ax = figure.add_axes([0, 0, 1, 1])

    # Plot the data.
    if isinstance(bandflux, np.ndarray):
        bandflux = {"lightcurve": bandflux}
    for name, curve in bandflux.items():
        if times is None:
            times = np.arange(len(curve))
        ax.plot(times, curve, marker="o", label=name)

    # Set the title and axis labels.
    if title is not None:
        ax.set_title(title)
    ax.set_xlabel("Time (days)")
    ax.set_ylabel("Flux")

    # Only include a legend if there are at least two curves.
    if len(bandflux) > 1:
        ax.legend()

    return ax


def plot_flux_spectrogram(flux_density, times=None, wavelengths=None, ax=None, figure=None, title=None):
    """Plot a spectrogram to visualize the fluxes.

    Parameters
    ----------
    flux_density : numpy.ndarray
        A length T x N matrix of SED values (in nJy), where  T is the number of time steps,
        and N is the number of wavelengths.
    times : numpy.ndarray or None, optional
        A length T matrix of the times, used for setting the x axis. If not
        provided, uses equal spaced ticks. None by default.
    wavelengths : numpy.ndarray or None, optional
        A length N matrix of the times, used for setting the y axis. If not
        provided, uses equal spaced ticks. None by default.
    ax : matplotlib.pyplot.Axes or None, optional
        Axes, None by default.
    figure : matplotlib.pyplot.Figure or None
        Figure, None by default.
    title : str or None, optional
        Title of the plot. None by default.

    Returns
    -------
    ax : matplotlib.pyplot.Axes
        The axes containing the plot.
    """
    # If no axes were given create them using either the given figure or
    # a newly created one (if no figure is given).
    if ax is None:
        if figure is None:
            figure = plt.figure()
        ax = figure.add_axes([0, 0, 1, 1])

    ax.imshow(flux_density.T, cmap="plasma", interpolation="nearest", aspect="auto")

    # Add title, axis labels, and correct ticks
    if title is None:
        ax.set_title("Flux Spectrogram")
    else:
        ax.set_title(title)

    if times is not None:
        ax.set_xlabel("Time (days)")
        ax.set_xticks(np.arange(len(times))[::4], [f"{round(time)}" for time in times][::4])
    if wavelengths is not None:
        ax.set_ylabel("Wavelength (Angstrom)")
        ax.set_yticks(np.arange(len(wavelengths))[::50], [f"{round(wave)}" for wave in wavelengths][::50])

    # Add flux labels
    for (j, i), label in np.ndenumerate(flux_density.T):
        if i % 2 == 1 and j % 40 == 20:
            ax.text(i, j, round(label, 1), ha="center", va="center", size=8)

    return ax



import numpy as np
import pytest
from astropy import units as u
from lightcurvelynx.astro_utils.unit_utils import flam_to_fnu, fnu_to_flam


def test_flam_to_fnu():
    """Test flam to fnu conversion:
    fnu = flam * lam**2 / c
    """
    c = 299_792_458.0 * 1.0e10  # in AA/s
    wave = [100.0, 1.0e5]
    flam = [c * 1.0e-4, 1.0e5]
    fnu = flam_to_fnu(
        flam,
        wave,
        wave_unit=u.AA,
        flam_unit=u.erg / u.second / u.cm**2 / u.AA,
        fnu_unit=u.erg / u.second / u.cm**2 / u.Hz,
    )
    assert np.allclose(fnu, [1.0, 1.0e15 / c])


def test_flam_to_fnu_matrix():
    """Test flam to fnu conversion when dealing with a matrix of observations."""
    c = 299_792_458.0 * 1.0e10  # in AA/s

    # Both wave and flam are matrices
    fnu = flam_to_fnu(
        [[c * 1.0e-4, 1.0e5], [1000.0, 2000.0], [c * 1.0e-4, c * 1.0e-4]],
        [[100.0, 1.0e5], [1000.0, 1.0e4], [1000.0, 1000.0]],
        wave_unit=u.AA,
        flam_unit=u.erg / u.second / u.cm**2 / u.AA,
        fnu_unit=u.erg / u.second / u.cm**2 / u.Hz,
    )

    # Compute expected from fnu = flam * lam**2 / c
    expected = np.array([[1.0, 1.0e15 / c], [1.0e9 / c, 2.0e11 / c], [100.0, 100.0]])
    assert np.allclose(fnu, expected)

    # Use a single vector of wavelengths.
    fnu = flam_to_fnu(
        [[c * 1.0e-4, 1.0e5], [1000.0, 2000.0], [c * 1.0e-4, c * 1.0e-4]],
        [100.0, 1.0e5],
        wave_unit=u.AA,
        flam_unit=u.erg / u.second / u.cm**2 / u.AA,
        fnu_unit=u.erg / u.second / u.cm**2 / u.Hz,
    )
    expected = np.array([[1.0, 1.0e15 / c], [1.0e7 / c, 2.0e13 / c], [1.0, 1.0e6]])
    assert np.allclose(fnu, expected)

    # Fail if the dimensions are not compatible.
    with pytest.raises(ValueError):
        _ = flam_to_fnu(
            [[c * 1.0e-4, 1.0e5], [1000.0, 2000.0], [c * 1.0e-4, c * 1.0e-4]],
            [[100.0, 1.0e5], [1000.0, 1.0e4]],
            wave_unit=u.AA,
            flam_unit=u.erg / u.second / u.cm**2 / u.AA,
            fnu_unit=u.erg / u.second / u.cm**2 / u.Hz,
        )

    with pytest.raises(ValueError):
        _ = flam_to_fnu(
            [[c * 1.0e-4, 1.0e5], [1000.0, 2000.0], [c * 1.0e-4, c * 1.0e-4]],
            [100.0, 200.0, 300.0],
            wave_unit=u.AA,
            flam_unit=u.erg / u.second / u.cm**2 / u.AA,
            fnu_unit=u.erg / u.second / u.cm**2 / u.Hz,
        )


def test_flam_to_fnu_to_flam():
    """Test that flam_to_fnu(fnu_to_flam) is the identity."""
    rng = np.random.default_rng(None)
    n = 100
    waves = rng.uniform(low=100.0, high=1.0e5, size=n)
    flam0 = rng.lognormal(mean=0.0, sigma=1.0, size=n)
    fnu = flam_to_fnu(
        flam0,
        waves,
        wave_unit=u.AA,
        flam_unit=u.erg / u.second / u.cm**2 / u.AA,
        fnu_unit=u.nJy,
    )
    flam1 = fnu_to_flam(
        fnu,
        waves,
        wave_unit=u.AA,
        flam_unit=u.erg / u.second / u.cm**2 / u.AA,
        fnu_unit=u.nJy,
    )
    np.testing.assert_allclose(flam0, flam1)

    # We throw an error on mismatched dimensions.
    with pytest.raises(ValueError):
        _ = fnu_to_flam(
            fnu,
            waves[1:],
            wave_unit=u.AA,
            flam_unit=u.erg / u.second / u.cm**2 / u.AA,
            fnu_unit=u.nJy,
        )



import numpy as np
import pytest
from lightcurvelynx.astro_utils.passbands import Passband, PassbandGroup
from lightcurvelynx.effects.basic_effects import ScaleFluxEffect
from lightcurvelynx.math_nodes.np_random import NumpyRandomFunc
from lightcurvelynx.models.basic_models import ConstantSEDModel, StepModel
from lightcurvelynx.models.multi_object_model import AdditiveMultiObjectModel, RandomMultiObjectModel
from lightcurvelynx.models.static_sed_model import StaticBandfluxModel, StaticSEDModel


def test_additive_multi_object_node() -> None:
    """Test that we can create and evaluate a AdditiveMultiObjectModel."""
    object1 = ConstantSEDModel(brightness=10.0, node_label="my_static_object")
    object2 = StepModel(brightness=15.0, t0=1.0, t1=2.0, node_label="my_step_object")
    model = AdditiveMultiObjectModel([object1, object2], node_label="my_multi_object")

    state = model.sample_parameters()
    assert state["my_static_object"]["brightness"] == 10.0
    assert state["my_step_object"]["brightness"] == 15.0

    times = np.array([0.0, 1.5, 3.0])
    wavelengths = np.array([1000.0, 2000.0])

    values = model.evaluate_sed(times, wavelengths, state)
    assert values.shape == (3, 2)
    assert np.allclose(values, [[10.0, 10.0], [25.0, 25.0], [10.0, 10.0]])


def test_additive_multi_object_node_passband() -> None:
    """Test that we can evaluate a AdditiveMultiObjectModel at the passband level."""
    a_band = Passband(np.array([[900, 0.0], [1000, 0.5], [2000, 0.5], [2100, 0.0]]), "LSST", "a")
    b_band = Passband(np.array([[2900, 0.0], [3000, 0.5], [4000, 0.5], [4100, 0.0]]), "LSST", "b")
    c_band = Passband(np.array([[5900, 0.0], [6000, 0.5], [7000, 0.5], [7100, 0.0]]), "LSST", "c")
    pb_group = PassbandGroup([a_band, b_band, c_band])

    object1 = ConstantSEDModel(brightness=10.0, node_label="my_static_object")
    object2 = StepModel(brightness=15.0, t0=1.0, t1=2.0, node_label="my_step_object")
    model = AdditiveMultiObjectModel([object1, object2], node_label="my_multi_object")
    state = model.sample_parameters(num_samples=1)

    times = np.array([0.0, 0.5, 1.25, 1.5, 3.0, 4.0])
    filters = np.array(["a", "a", "a", "b", "c", "a"])

    bandflux1 = object1.evaluate_bandfluxes(pb_group, times, filters, state)
    assert np.allclose(bandflux1, [10.0, 10.0, 10.0, 10.0, 10.0, 10.0])

    bandflux2 = object2.evaluate_bandfluxes(pb_group, times, filters, state)
    assert np.allclose(bandflux2, [0.0, 0.0, 15.0, 15.0, 0.0, 0.0])

    bandflux_combined = model.evaluate_bandfluxes(pb_group, times, filters, state)
    assert np.allclose(bandflux_combined, [10.0, 10.0, 25.0, 25.0, 10.0, 10.0])

    # We can include a Bandflux only model in the computation.
    object3 = StaticBandfluxModel({"a": 1.0, "b": 2.0, "c": 0.0})
    model2 = AdditiveMultiObjectModel([object1, object2, object3], node_label="my_multi_object")
    state2 = model2.sample_parameters(num_samples=1)

    bandflux_combined = model2.evaluate_bandfluxes(pb_group, times, filters, state2)
    assert np.allclose(bandflux_combined, [11.0, 11.0, 26.0, 27.0, 10.0, 11.0])


def test_additive_multi_object_node_resample() -> None:
    """Test that we can correctly resample a AdditiveMultiObjectModel."""
    ra = NumpyRandomFunc("uniform", low=10.0, high=20.0)
    dec = NumpyRandomFunc("uniform", low=-10.0, high=10.0)

    # object1 and object2 share the same ra and dec for each sample, but have
    # different brightness values.
    object1 = ConstantSEDModel(
        brightness=NumpyRandomFunc("uniform", low=10.0, high=20.0),
        ra=ra,
        dec=dec,
        node_label="my_static_object",
    )
    object2 = StepModel(
        brightness=NumpyRandomFunc("uniform", low=30.0, high=40.0),
        ra=ra,
        dec=dec,
        t0=1.0,
        t1=2.0,
        node_label="my_step_object",
    )
    model = AdditiveMultiObjectModel([object1, object2], node_label="my_multi_object")

    num_samples = 1000
    rng_info = np.random.default_rng(100)
    state = model.sample_parameters(num_samples=num_samples, rng_info=rng_info)
    assert len(np.unique(state["my_static_object"]["ra"])) > num_samples / 2
    assert len(np.unique(state["my_static_object"]["dec"])) > num_samples / 2
    assert np.allclose(state["my_static_object"]["ra"], state["my_static_object"]["ra"])
    assert np.allclose(state["my_static_object"]["dec"], state["my_static_object"]["dec"])

    brightness_diff = state["my_step_object"]["brightness"] - state["my_static_object"]["brightness"]
    assert np.all(brightness_diff > 5.0)

    # We can evaluate the combined model at all of the times.
    times = np.array([0.0, 1.5, 3.0])
    wavelengths = np.array([1000.0, 2000.0])

    values = model.evaluate_sed(times, wavelengths, state)
    assert values.shape == (1000, 3, 2)

    assert np.allclose(values[:, 0, 0], state["my_static_object"]["brightness"])
    assert np.allclose(values[:, 0, 1], state["my_static_object"]["brightness"])
    assert np.allclose(
        values[:, 1, 0],
        state["my_static_object"]["brightness"] + state["my_step_object"]["brightness"],
    )
    assert np.allclose(
        values[:, 1, 1],
        state["my_static_object"]["brightness"] + state["my_step_object"]["brightness"],
    )
    assert np.allclose(values[:, 2, 0], state["my_static_object"]["brightness"])
    assert np.allclose(values[:, 2, 1], state["my_static_object"]["brightness"])


def test_additive_multi_object_node_redshift() -> None:
    """Test that we handle redshifts separately for each object."""
    object1 = StepModel(brightness=10.0, t0=1.0, t1=3.0, redshift=0.0, node_label="object1")
    object2 = StepModel(brightness=10.0, t0=2.0, t1=4.0, redshift=1.0, node_label="object2")
    model = AdditiveMultiObjectModel([object1, object2], node_label="my_multi_object")

    state = model.sample_parameters()
    times = np.array([0.5, 1.5, 2.5, 3.5, 4.5])
    wavelengths = np.array([1000.0, 2000.0])

    # Compute the expected values:
    # The redshift applied to object1 is a no-op for both time and brightness.
    # The redshift applied to object2 shifts times to [1.25 1.75 2.25 2.75 3.25]
    # and scales up the brightness by a factor of (1 + z) = 2.0.
    contrib1 = np.array([[0.0, 0.0], [10.0, 10.0], [10.0, 10.0], [0.0, 0.0], [0.0, 0.0]])
    contrib2 = np.array([[0.0, 0.0], [0.0, 0.0], [20.0, 20.0], [20.0, 20.0], [20.0, 20.0]])

    values = model.evaluate_sed(times, wavelengths, state)
    assert values.shape == (5, 2)
    assert np.allclose(values, contrib1 + contrib2)


def test_additive_multi_object_node_effects_rest_frame() -> None:
    """Test that we handle rest frame effects separately for each object."""
    object1 = StepModel(brightness=10.0, t0=1.0, t1=3.0, node_label="object1")
    object2 = StepModel(brightness=10.0, t0=2.0, t1=4.0, node_label="object2")
    object1.add_effect(ScaleFluxEffect(flux_scale=0.5, rest_frame=True))
    object2.add_effect(ScaleFluxEffect(flux_scale=0.1, rest_frame=True))
    model = AdditiveMultiObjectModel([object1, object2], node_label="my_multi_object")

    # We added the effect to each submodel's rest frame list.
    assert len(object1.rest_frame_effects) == 1
    assert len(object2.rest_frame_effects) == 1
    assert len(object1.obs_frame_effects) == 0
    assert len(object2.obs_frame_effects) == 0
    assert len(model.rest_frame_effects) == 0
    assert len(model.obs_frame_effects) == 0

    state = model.sample_parameters()
    times = np.array([0.5, 1.5, 2.5, 3.5, 4.5])
    wavelengths = np.array([1000.0, 2000.0])
    values = model.evaluate_sed(times, wavelengths, state)
    assert values.shape == (5, 2)

    contrib1 = np.array([[0.0, 0.0], [5.0, 5.0], [5.0, 5.0], [0.0, 0.0], [0.0, 0.0]])
    contrib2 = np.array([[0.0, 0.0], [0.0, 0.0], [1.0, 1.0], [1.0, 1.0], [0.0, 0.0]])
    assert np.allclose(values, contrib1 + contrib2)


def test_additive_multi_object_node_effects_rest_frame_add() -> None:
    """Test that we handle rest frame effects separately for each object."""
    object1 = StepModel(brightness=10.0, t0=1.0, t1=3.0, node_label="object1")
    object2 = StepModel(brightness=10.0, t0=2.0, t1=4.0, node_label="object2")
    model = AdditiveMultiObjectModel([object1, object2], node_label="my_multi_object")
    model.add_effect(ScaleFluxEffect(flux_scale=0.5, rest_frame=True))

    # We added the effect to each submodel's rest frame list.
    assert len(object1.rest_frame_effects) == 1
    assert len(object2.rest_frame_effects) == 1
    assert len(object1.obs_frame_effects) == 0
    assert len(object2.obs_frame_effects) == 0
    assert len(model.rest_frame_effects) == 0
    assert len(model.obs_frame_effects) == 0

    state = model.sample_parameters()
    times = np.array([0.5, 1.5, 2.5, 3.5, 4.5])
    wavelengths = np.array([1000.0, 2000.0])
    values = model.evaluate_sed(times, wavelengths, state)
    assert values.shape == (5, 2)

    contrib1 = np.array([[0.0, 0.0], [5.0, 5.0], [5.0, 5.0], [0.0, 0.0], [0.0, 0.0]])
    contrib2 = np.array([[0.0, 0.0], [0.0, 0.0], [5.0, 5.0], [5.0, 5.0], [0.0, 0.0]])
    assert np.allclose(values, contrib1 + contrib2)


def test_additive_multi_object_node_effects_obs_frame() -> None:
    """Test that we handle observer frame effects together."""
    object1 = StepModel(brightness=10.0, t0=1.0, t1=3.0, node_label="object1")
    object2 = StepModel(brightness=10.0, t0=2.0, t1=4.0, node_label="object2")
    model = AdditiveMultiObjectModel([object1, object2], node_label="my_multi_object")
    model.add_effect(ScaleFluxEffect(flux_scale=0.5, rest_frame=False))

    # We added the effect to the joint model's rest frame list.
    assert len(object1.rest_frame_effects) == 0
    assert len(object2.rest_frame_effects) == 0
    assert len(object1.obs_frame_effects) == 0
    assert len(object2.obs_frame_effects) == 0
    assert len(model.rest_frame_effects) == 0
    assert len(model.obs_frame_effects) == 1

    state = model.sample_parameters()
    times = np.array([0.5, 1.5, 2.5, 3.5, 4.5])
    wavelengths = np.array([1000.0, 2000.0])
    values = model.evaluate_sed(times, wavelengths, state)
    assert values.shape == (5, 2)

    contrib1 = np.array([[0.0, 0.0], [5.0, 5.0], [5.0, 5.0], [0.0, 0.0], [0.0, 0.0]])
    contrib2 = np.array([[0.0, 0.0], [0.0, 0.0], [5.0, 5.0], [5.0, 5.0], [0.0, 0.0]])
    assert np.allclose(values, contrib1 + contrib2)


def test_additive_multi_object_node_min_max() -> None:
    """Test that we can get the correct wavelength limits for a AdditiveMultiObjectModel."""
    sed0 = np.array(
        [
            [100.0, 200.0, 300.0, 400.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
        ]
    )
    model0 = StaticSEDModel([sed0], node_label="sed0")

    sed1 = np.array(
        [
            [200.0, 300.0, 400.0, 500.0],  # Wavelengths
            [20.0, 40.0, 40.0, 20.0],  # fluxes
        ]
    )
    model1 = StaticSEDModel([sed1], node_label="sed1")

    # The reported min/max wavelengths are the overlap of the objects.
    model = AdditiveMultiObjectModel(
        [model0, model1],
        node_label="test",
    )
    states = model.sample_parameters(num_samples=1)
    assert np.array_equal(model.minwave(states), [100.0, 200.0])
    assert np.array_equal(model.maxwave(states), [400.0, 500.0])


def test_additive_multi_object_node_bandflux() -> None:
    """Test that we can create and evaluate an AdditiveMultiObjectModel with Bandflux models."""
    object1 = StaticBandfluxModel({"a": 1.0, "b": 2.0, "c": 0.0}, node_label="object1")
    object2 = StaticBandfluxModel({"a": 1.0, "b": 1.0, "c": 0.5}, node_label="object2")
    model = AdditiveMultiObjectModel([object1, object2], node_label="test")

    # When we evaluate the model, we should get the expected values.
    state = model.sample_parameters()
    times = np.array([0.0, 1.5, 3.0, 4.0, 5.0])
    filters = np.array(["a", "a", "b", "a", "c"])
    values = model.evaluate_bandfluxes(None, times, filters, state)
    assert np.array_equal(values, [2.0, 2.0, 3.0, 2.0, 0.5])

    # Check that we cannot turn on "apply_redshift" when using bandflux models.
    with pytest.raises(NotImplementedError):
        model.set_apply_redshift(True)

    # Check that we can add a rest frame effect. This are applied at the bandpass
    # level for Bandpass models.
    model.add_effect(ScaleFluxEffect(flux_scale=0.5, rest_frame=True))
    state = model.sample_parameters()
    values = model.evaluate_bandfluxes(None, times, filters, state)
    assert np.allclose(values, [1.0, 1.0, 1.5, 1.0, 0.25])

    # Check that we can apply a observer frame effect. The two effects stack.
    model.add_effect(ScaleFluxEffect(flux_scale=0.1, rest_frame=False))
    state = model.sample_parameters()
    values = model.evaluate_bandfluxes(None, times, filters, state)
    assert np.allclose(values, [0.1, 0.1, 0.15, 0.1, 0.025])

    # Check that we fail to evaluate the SEDs when we have a bandflux model.
    with pytest.raises(TypeError):
        _ = model.evaluate_sed(times, np.array([4000.0, 5000.0]), state)


def test_random_multi_object_node() -> None:
    """Test that we can create and evaluate a RandomMultiObjectModel."""
    object1 = ConstantSEDModel(brightness=10.0, node_label="object1")
    object2 = ConstantSEDModel(brightness=15.0, node_label="object2")
    model = RandomMultiObjectModel(
        [object1, object2],
        weights=[0.8, 0.2],
        node_label="my_multi_object",
    )

    state = model.sample_parameters(num_samples=10_000)
    assert np.all(state["object1"]["brightness"] == 10.0)
    assert np.all(state["object2"]["brightness"] == 15.0)

    # We should get approximately 80% of the samples from the first
    # object and 20% from the second object.
    object = np.array(state["my_multi_object"]["selected_object"], dtype=str)
    assert np.all((object == "object1") | (object == "object2"))
    assert np.sum(object == "object1") > 7000
    assert np.sum(object == "object2") > 1000

    # When we evaluate the model, we should get the expected values.
    times = np.array([0.0, 1.5, 3.0])
    wavelengths = np.array([1000.0, 2000.0])
    values = model.evaluate_sed(times, wavelengths, state)

    assert values.shape == (10_000, 3, 2)
    assert np.all((values == 10.0) | (values == 15.0))
    assert np.sum(values == 10.0) > 7000 * 5
    assert np.sum(values == 15.0) > 1000 * 5


def test_random_multi_object_node_bandflux() -> None:
    """Test that we can create and evaluate a RandomMultiObjectModel."""
    a_band = Passband(np.array([[900, 0.0], [1000, 0.5], [2000, 0.5], [2100, 0.0]]), "LSST", "a")
    b_band = Passband(np.array([[2900, 0.0], [3000, 0.5], [4000, 0.5], [4100, 0.0]]), "LSST", "b")
    c_band = Passband(np.array([[5900, 0.0], [6000, 0.5], [7000, 0.5], [7100, 0.0]]), "LSST", "c")
    pb_group = PassbandGroup([a_band, b_band, c_band])

    # Create a random model with 2 SED-based models and 1-bandflux based model.
    object1 = ConstantSEDModel(brightness=10.0, node_label="object1")
    object2 = ConstantSEDModel(brightness=15.0, node_label="object2")
    object3 = StaticBandfluxModel({"a": 1.0, "b": 2.0, "c": 0.0}, node_label="object3")
    model = RandomMultiObjectModel(
        [object1, object2, object3],
        weights=[0.25, 0.5, 0.25],
        node_label="my_multi_object",
    )

    rng_info = np.random.default_rng(100)
    state = model.sample_parameters(num_samples=1_000, rng_info=rng_info)

    model_names = np.array(state["my_multi_object"]["selected_object"], dtype=str)
    assert np.all((model_names == "object1") | (model_names == "object2") | (model_names == "object3"))
    assert np.sum(model_names == "object1") > 200
    assert np.sum(model_names == "object2") > 400
    assert np.sum(model_names == "object3") > 200

    # When we evaluate the model, we should get the expected values.
    times = np.array([0.0, 1.5, 3.0, 4.0])
    filters = np.array(["a", "a", "b", "a"])
    values = model.evaluate_bandfluxes(pb_group, times, filters, state)

    assert values.shape == (1_000, 4)
    for i in range(1_000):
        if model_names[i] == "object1":
            assert np.allclose([10.0, 10.0, 10.0, 10.0], values[i])
        elif model_names[i] == "object2":
            assert np.allclose([15.0, 15.0, 15.0, 15.0], values[i])
        elif model_names[i] == "object3":
            assert np.allclose([1.0, 1.0, 2.0, 1.0], values[i])

    # Check that we fail if we try to evaluate the SEDs.
    with pytest.raises(TypeError):
        _ = model.evaluate_sed(times, np.array([1000.0, 2000.0]), state)


def test_random_multi_object_node_min_max() -> None:
    """Test that we can get the correct wavelength limits for a RandomMultiObjectModel."""
    sed0 = np.array(
        [
            [100.0, 200.0, 300.0, 400.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
        ]
    )
    model0 = StaticSEDModel([sed0], node_label="sed0")

    sed1 = np.array(
        [
            [200.0, 300.0, 400.0, 500.0],  # Wavelengths
            [20.0, 40.0, 40.0, 20.0],  # fluxes
        ]
    )
    model1 = StaticSEDModel([sed1], node_label="sed1")

    model = RandomMultiObjectModel(
        [model0, model1],
        weights=[0.5, 0.5],
        node_label="test",
    )
    states = model.sample_parameters(num_samples=1)

    # Force the selected_object to be 0 for the test.
    states.set("test", "selected_object", 0)
    assert model.minwave(states) == 100.0
    assert model.maxwave(states) == 400.0

    # Force the selected_object to be 1 for the test.
    states.set("test", "selected_object", 1)
    assert model.minwave(states) == 200.0
    assert model.maxwave(states) == 500.0

    # We fail if we do not pass in the states.
    with pytest.raises(ValueError):
        _ = model.minwave()
    with pytest.raises(ValueError):
        _ = model.maxwave()



from pathlib import Path
from unittest.mock import patch

import numpy as np
import pytest
from astropy import units as u
from lightcurvelynx import _LIGHTCURVELYNX_TEST_DATA_DIR
from lightcurvelynx.astro_utils.unit_utils import fnu_to_flam
from lightcurvelynx.math_nodes.np_random import NumpyRandomFunc
from lightcurvelynx.models.sncomso_models import SncosmoWrapperModel
from lightcurvelynx.utils.extrapolate import ExponentialDecay


def _fake_nugent_data_path(*args, **kwargs):
    """A function for pointing to the test data directory's version of the nugent model file."""
    return str(
        Path(_LIGHTCURVELYNX_TEST_DATA_DIR) / "fake_sncosmo" / "models" / "nugent" / "sn1a_flux.v1.2.dat"
    )


def test_sncomso_models_hsiao() -> None:
    """Test that we can create and evalue a 'hsiao' model."""
    model = SncosmoWrapperModel("hsiao", t0=0.0, amplitude=2.0e10)
    state = model.sample_parameters()
    assert model.get_param(state, "amplitude") == 2.0e10
    assert model.get_param(state, "t0") == 0.0
    assert str(model) == "SncosmoWrapperModel_0"

    assert np.array_equal(model.param_names, ["amplitude"])
    assert np.array_equal(model.parameter_values, [2.0e10])

    # Test against a manual query from sncosmo with no redshift and an
    # unrealistically high amplitude.
    #     model = sncosmo.Model(source='hsiao')
    #     model.set(z=0.0, t0=0.0, amplitude=2.0e10)
    #     model.flux(5., [4000., 4100., 4200.])
    fluxes_fnu = model.evaluate_sed([5.0], [4000.0, 4100.0, 4200.0])
    fluxes_flam = fnu_to_flam(
        fluxes_fnu,
        [4000.0, 4100.0, 4200.0],
        wave_unit=u.AA,
        flam_unit=u.erg / u.second / u.cm**2 / u.AA,
        fnu_unit=u.nJy,
    )
    assert np.allclose(fluxes_flam, [133.98143039, 152.74613574, 134.40916824])

    # Check that we can mask times.  The 'hsiao' model uses phases (-20.0, 85.0).
    sample_times = np.arange(-50.0, 100.0, 1.0)
    mask = model.mask_by_time(sample_times, graph_state=state)

    expected_mask = (sample_times > -20.0) & (sample_times < 85.0)
    assert np.array_equal(mask, expected_mask)


def test_sncomso_models_hsiao_t0() -> None:
    """Test that we can create and evalue a 'hsiao' model with a t0."""
    model = SncosmoWrapperModel("hsiao", t0=55000.0, amplitude=2.0e10)
    state = model.sample_parameters()
    assert model.get_param(state, "amplitude") == 2.0e10
    assert model.get_param(state, "t0") == 55000.0

    assert np.array_equal(model.param_names, ["amplitude"])
    assert np.array_equal(model.parameter_values, [2.0e10])

    # Test against a manual query from sncosmo with no redshift and an unrealistically high amplitude.
    #     model = sncosmo.Model(source='hsiao')
    #     model.set(z=0.0, t0=55000., amplitude=2.0e10)
    #     model.flux(54990., [4000., 4100., 4200.])
    fluxes_fnu = model.evaluate_sed([54990.0], [4000.0, 4100.0, 4200.0])
    fluxes_flam = fnu_to_flam(
        fluxes_fnu,
        [4000.0, 4100.0, 4200.0],
        wave_unit=u.AA,
        flam_unit=u.erg / u.second / u.cm**2 / u.AA,
        fnu_unit=u.nJy,
    )
    assert np.allclose(fluxes_flam, [67.83696271, 67.98471119, 47.20395186])

    # Check that we can mask times.  The 'hsiao' model uses phases (-20.0, 85.0),
    # which is offset by t0=55000.0.
    sample_times = np.arange(-50.0, 100.0, 1.0) + 55000.0
    mask = model.mask_by_time(sample_times, graph_state=state)

    expected_mask = (sample_times > 54980.0) & (sample_times < 55085.0)
    assert np.array_equal(mask, expected_mask)

    # We raise a warning if the time is outside the bounds.
    with np.testing.assert_warns(UserWarning):
        model.evaluate_sed([0.0], [4000.0, 4100.0])


def test_sncomso_models_hsiao_extrap_time() -> None:
    """Test that we can extrapolate to times outside the model bounds."""
    time_extrapolation = ExponentialDecay(rate=0.1)
    model = SncosmoWrapperModel(
        "hsiao",
        t0=55000.0,
        amplitude=2.0e10,
        time_extrapolation=time_extrapolation,
    )

    query_times = np.array([54500.0, 54900.0, 54960.0, 54990.0, 55000.0, 55010.0, 55100.0, 55620.0, 99990.0])
    fluxes = model.evaluate_sed(query_times, [4000.0])

    # All the times before are monotonically increasing to the first valid time.
    assert np.all(fluxes[0:3] < fluxes[3])
    assert np.all(np.diff(fluxes[0:3]) >= 0.0)
    # All the times after are monotonically decreasing from the last valid time.

    assert np.all(fluxes[7:9] < fluxes[6])
    assert np.all(np.diff(fluxes[5:9]) <= 0.0)


def test_sncomso_models_bounds() -> None:
    """Test that we do not crash if we give wavelengths outside the model bounds."""
    # Use a massively subsampled version of the 'nugent-sn1a' model (only 3 time steps)
    # that is cached in the test data directory. This is okay because we are only using
    # the model's wavelength bounds.
    with patch("sncosmo.utils.DataMirror.abspath", side_effect=_fake_nugent_data_path):
        model = SncosmoWrapperModel("nugent-sn1a", amplitude=2.0e10, t0=54990.0)
    min_w = model.source.minwave()
    max_w = model.source.maxwave()

    wavelengths = [
        min_w - 100.0,  # Out of bounds
        min_w,  # edge of bounds (included)
        0.5 * min_w + 0.5 * max_w,  # included
        max_w,  # edge of bounds (included)
        max_w + 0.1,  # Out of bounds
    ]

    # SNCosmo raises an error if we give wavelengths outside the bounds. It should
    # raise a warning first.
    with pytest.warns(UserWarning):
        with pytest.raises(ValueError):
            _ = model.evaluate_sed([54990.0, 54990.5], wavelengths)


def test_sncomso_models_linear_extrapolate() -> None:
    """Test that we do not crash if we give wavelengths outside the model bounds."""
    # Use a massively subsampled version of the 'nugent-sn1a' model (only 3 time steps)
    # that is cached in the test data directory. This is okay because we are only using
    # the model's wavelength bounds.
    with patch("sncosmo.utils.DataMirror.abspath", side_effect=_fake_nugent_data_path):
        model = SncosmoWrapperModel(
            "nugent-sn1a",
            amplitude=2.0e10,
            t0=54990.0,
            wave_extrapolation=ExponentialDecay(rate=0.1),
        )
    min_w = model.source.minwave()
    max_w = model.source.maxwave()

    wavelengths = [
        min_w - 100.0,  # Out of bounds
        min_w,  # edge of bounds (included)
        0.5 * min_w + 0.5 * max_w,  # included
        max_w,  # edge of bounds (included)
        max_w + 0.1,  # Out of bounds
        max_w + 500.0,  # Out of bounds
        max_w + 5000.0,  # Out of bounds
    ]

    # Check that columns 0, 4, 5, and 6 are correctly extrapolated.
    fluxes_fnu = model.evaluate_sed([54990.0, 54990.5], wavelengths)
    assert np.all(fluxes_fnu[:, 0:6] > 0.0)
    assert np.all(fluxes_fnu[:, 0] < fluxes_fnu[:, 1])
    assert np.all(fluxes_fnu[:, 4] < fluxes_fnu[:, 3])
    assert np.all(fluxes_fnu[:, 5] < fluxes_fnu[:, 4])
    assert np.all(fluxes_fnu[:, 6] < fluxes_fnu[:, 5])


def test_sncomso_models_set() -> None:
    """Test that we can create and evalue a 'hsiao' model and set parameter."""
    model = SncosmoWrapperModel("hsiao", t0=0.0, redshift=0.5)

    # sncosmo parameters exist at default values.
    assert np.array_equal(model.param_names, ["amplitude"])
    assert np.array_equal(model.parameter_values, [1.0])

    model.set(amplitude=100.0)
    state = model.sample_parameters()

    assert model.get_param(state, "amplitude") == 100.0
    assert np.array_equal(model.param_names, ["amplitude"])
    assert np.array_equal(model.parameter_values, [100.0])


def test_sncomso_models_chained() -> None:
    """Test that we can create and evalue a 'hsiao' model using randomized parameters."""
    # Generate the amplitude from a uniform distribution, but use a fixed seed so we have
    # reproducible tests.
    model = SncosmoWrapperModel(
        "hsiao",
        amplitude=NumpyRandomFunc("uniform", low=2.0, high=12.0, seed=100),
    )
    _ = model.sample_parameters()

    assert np.array_equal(model.param_names, ["amplitude"])
    assert 2.0 <= model.parameter_values[0] <= 12.0

    # When we resample, we get different model parameters. Create a histogram
    # of 10,000 samples.
    hist = [0] * 10
    source_model = model.source
    for _ in range(10_000):
        _ = model.sample_parameters()
        bin = int(source_model.parameters[0] - 2.0)
        hist[bin] += 1

    # Each bin should have around 1,000 samples.
    for i in range(10):
        assert 800 < hist[i] < 1200



import matplotlib

matplotlib.use("Agg")  # Suppress the plots

import numpy as np
import pytest
from astropy.table import Table
from citation_compass import find_in_citations
from lightcurvelynx.astro_utils.mag_flux import mag2flux
from lightcurvelynx.astro_utils.passbands import Passband, PassbandGroup
from lightcurvelynx.effects.basic_effects import ScaleFluxEffect
from lightcurvelynx.models.lightcurve_template_model import (
    LightcurveBandData,
    LightcurveTemplateModel,
    MultiLightcurveTemplateModel,
)
from lightcurvelynx.utils.extrapolate import LinearDecay


def _create_toy_passbands() -> PassbandGroup:
    """Create a toy passband group with three passbands where the first passband
    has no overlap while the second two overlap each other for half the range.
    """
    a_band = Passband(np.array([[400, 0.5], [500, 0.5], [600, 0.5]]), "LSST", "u")
    b_band = Passband(np.array([[800, 0.8], [900, 0.8], [1000, 0.8]]), "LSST", "g")
    c_band = Passband(np.array([[900, 0.6], [1000, 0.6], [1100, 0.6]]), "LSST", "r")
    return PassbandGroup(given_passbands=[a_band, b_band, c_band])


def _create_toy_lightcurves() -> dict:
    """Create toy light curves where the first two are constant and the third
    is linearly increasing.  Each light curve covers a slightly different
    time range.
    """
    times = np.linspace(1, 11, 20)
    lightcurves = {
        "u": np.array([times - 0.2, 2.0 * np.ones_like(times)]).T,
        "g": np.array([times - 0.1, 3.0 * np.ones_like(times)]).T,
        "r": np.array([times, 0.1 * times + 1.0]).T,
    }
    return lightcurves


def test_create_lightcurve_band_data_from_dict() -> None:
    """Test that we can create a simple LightcurveBandData object from a dict."""
    lightcurves = _create_toy_lightcurves()
    lc_data = LightcurveBandData(lightcurves, lc_data_t0=0.0)

    # Check the internal structure of the LightcurveBandData.
    assert len(lc_data) == 3
    assert lc_data.lc_data_t0 == 0.0
    assert lc_data.period is None
    assert lc_data.filters == ["u", "g", "r"]
    for filt in ["u", "g", "r"]:
        assert np.allclose(lc_data.lightcurves[filt], lightcurves[filt])
        assert lc_data.baseline[filt] == 0.0
    assert lc_data.has_time_bounds
    assert lc_data.min_times == {"u": 0.8, "g": 0.9, "r": 1.0}
    assert lc_data.max_times == {"u": 10.8, "g": 10.9, "r": 11.0}

    # If we use an lc_data_t0, we should shift the light curves' times accordingly
    # and provide a baseline.
    lc_data2 = LightcurveBandData(lightcurves, lc_data_t0=2.0, baseline={"u": 0.1, "g": 0.2, "r": 0.3})
    assert lc_data2.lc_data_t0 == 2.0
    assert lc_data2.period is None
    for filt in ["u", "g", "r"]:
        assert np.allclose(lc_data2.lightcurves[filt][:, 0], lightcurves[filt][:, 0] - 2.0)
        assert np.allclose(lc_data2.lightcurves[filt][:, 1], lightcurves[filt][:, 1])
    assert lc_data2.baseline == {"u": 0.1, "g": 0.2, "r": 0.3}
    assert not lc_data2.has_time_bounds  # baseline provided

    # We fail if the baseline does not match the filters.
    with pytest.raises(ValueError):
        _ = LightcurveBandData(lightcurves, lc_data_t0=2.0, baseline={"u": 0.1, "g": 0.2, "i": 0.3})

    # If we mark them as periodic we should fail because the times do not match.
    with pytest.raises(ValueError):
        _ = LightcurveBandData(lightcurves, lc_data_t0=0.0, periodic=True)

    # Check that we can specify the light curves in magnitudes.
    lc_data3 = LightcurveBandData(
        lightcurves,
        lc_data_t0=0.0,
        magnitudes_in=True,
        baseline={"u": 0.1, "g": 0.2, "r": 0.3},
    )
    assert len(lc_data3) == 3
    assert lc_data3.filters == ["u", "g", "r"]
    for filt in ["u", "g", "r"]:
        assert np.allclose(lc_data3.lightcurves[filt][:, 0], lightcurves[filt][:, 0])
        assert np.allclose(lc_data3.lightcurves[filt][:, 1], mag2flux(lightcurves[filt][:, 1]))
    assert lc_data3.baseline == {"u": mag2flux(0.1), "g": mag2flux(0.2), "r": mag2flux(0.3)}
    assert not lc_data3.has_time_bounds  # baseline provided

    # We fail if we try to create a light curve without a t0.
    with pytest.raises(ValueError):
        _ = LightcurveBandData(lightcurves, lc_data_t0=None)


def test_create_lightcurve_band_data_periodic_from_dict() -> None:
    """Test that we can create a periodic LightcurveBandData object from a dict."""
    times = np.linspace(3, 13, 20)
    lightcurves = {
        "u": np.array([times, 2.0 * np.ones_like(times)]).T,
        "g": np.array([times, 3.0 * np.ones_like(times)]).T,
    }
    lc_data = LightcurveBandData(lightcurves, lc_data_t0=0.0, periodic=True)

    assert len(lc_data) == 2
    assert lc_data.filters == ["u", "g"]
    assert lc_data.lc_data_t0 == 3.0
    assert lc_data.period == 10.0
    assert not lc_data.has_time_bounds  # periodic

    # The values should be the same and the times should be shifted by 3.0.
    for filt in ["u", "g"]:
        assert np.allclose(lc_data.lightcurves[filt][:, 0], lightcurves[filt][:, 0] - 3.0)
        assert np.allclose(lc_data.lightcurves[filt][:, 1], lightcurves[filt][:, 1])
    assert lc_data.baseline == {"u": 0.0, "g": 0.0}


def test_create_lightcurve_band_data_from_numpy() -> None:
    """Test that we can create a simple LightcurveBandData object from a numpy array."""
    lightcurves = np.array(
        [
            [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0],
            [1.0, 2.0, 1.0, 1.5, 2.0, 2.5, 1.5, 1.0, 0.0, 1.0],
            ["u", "g", "r", "u", "g", "r", "u", "g", "r", "u"],
        ]
    ).T
    lc_data = LightcurveBandData(lightcurves, lc_data_t0=0.0)

    # Check the internal structure of the LightcurveBandData.
    assert len(lc_data) == 3
    assert lc_data.lc_data_t0 == 0.0
    assert lc_data.period is None
    assert set(lc_data.filters) == {"u", "g", "r"}
    assert np.allclose(lc_data.lightcurves["u"][:, 0], [0.0, 3.0, 6.0, 9.0])
    assert np.allclose(lc_data.lightcurves["u"][:, 1], [1.0, 1.5, 1.5, 1.0])
    assert np.allclose(lc_data.lightcurves["g"][:, 0], [1.0, 4.0, 7.0])
    assert np.allclose(lc_data.lightcurves["g"][:, 1], [2.0, 2.0, 1.0])
    assert np.allclose(lc_data.lightcurves["r"][:, 0], [2.0, 5.0, 8.0])
    assert np.allclose(lc_data.lightcurves["r"][:, 1], [1.0, 2.5, 0.0])
    assert lc_data.baseline == {"u": 0.0, "g": 0.0, "r": 0.0}
    assert lc_data.has_time_bounds
    assert lc_data.min_times == {"u": 0.0, "g": 1.0, "r": 2.0}
    assert lc_data.max_times == {"u": 9.0, "g": 7.0, "r": 8.0}


def test_create_lightcurve_band_data_from_lclib_table() -> None:
    """Test that we can create a simple LightcurveBandData object from a LCLIB table."""
    # When creating the LightcurveBandData from a table, we specifc the values in magnitude,
    # instead of flux.
    data = {
        "time": [0.0, 1.0, 2.0, 3.0, 4.0, 0.0],
        "type": ["S", "S", "S", "S", "S", "T"],
        "u": [1.0, 2.0, 1.0, 1.5, 2.0, 1.0],
        "g": [2.0, 3.0, 2.0, 2.5, 3.0, 2.0],
        "r": [1.0, 1.5, 2.0, 2.5, 3.0, 2.0],
        "i": [0.5, 0.6, 0.7, 0.8, 0.9, 1.5],
    }
    table = Table(data)
    table.meta["RECUR_CLASS"] = "NON-RECUR"
    lc_data = LightcurveBandData.from_lclib_table(table)

    # Check the internal structure of the LightcurveBandData.
    assert len(lc_data) == 4
    assert lc_data.lc_data_t0 == 0.0
    assert lc_data.period is None
    assert lc_data.filters == ["u", "g", "r", "i"]

    for filt in ["u", "g", "r", "i"]:
        assert np.allclose(lc_data.lightcurves[filt][:, 0], data["time"][:5])
        assert np.allclose(
            lc_data.lightcurves[filt][:, 1],
            mag2flux(np.array(data[filt][:5])),
        )

    # We get the baseline from the "T" type row.
    print(lc_data.baseline)
    assert lc_data.baseline["u"] == mag2flux(1.0)
    assert lc_data.baseline["g"] == mag2flux(2.0)
    assert lc_data.baseline["r"] == mag2flux(2.0)
    assert lc_data.baseline["i"] == mag2flux(1.5)

    # If we specify a subset of filters, only load those.
    lc_data = LightcurveBandData.from_lclib_table(table, filters=["u", "g"])
    assert set(lc_data.filters) == set(["u", "g"])
    assert "r" not in lc_data.lightcurves
    assert "i" not in lc_data.lightcurves
    assert "u" in lc_data.lightcurves
    assert "g" in lc_data.lightcurves


def test_create_lightcurve_band_data_from_lclib_table_times() -> None:
    """Test that we can create a simple LightcurveBandData object from a LCLIB table
    with non-zero starting times."""
    # When creating the LightcurveBandData from a table, we specifc the values in magnitude,
    # instead of flux.
    data = {
        "time": [5.0, 6.0, 7.0, 8.0, 9.0],
        "type": ["S", "S", "S", "S", "S"],
        "u": [1.0, 2.0, 1.0, 1.5, 2.0],
        "g": [2.0, 3.0, 2.0, 2.5, 3.0],
    }
    table = Table(data)
    table.meta["RECUR_CLASS"] = "NON-RECUR"
    lc_data = LightcurveBandData.from_lclib_table(table)

    # Check the internal structure of the LightcurveBandData. And that we automatically zero shift
    # things from an LCLIB table.
    assert len(lc_data) == 2
    assert lc_data.lc_data_t0 == 5.0
    assert lc_data.filters == ["u", "g"]
    assert np.allclose(lc_data.lightcurves["u"][:, 0], [0.0, 1.0, 2.0, 3.0, 4.0])
    assert np.allclose(lc_data.lightcurves["g"][:, 0], [0.0, 1.0, 2.0, 3.0, 4.0])

    # Check that we can override the default zero-shifting behavior.
    lc_data = LightcurveBandData.from_lclib_table(table, forced_lc_t0=3.0)
    assert len(lc_data) == 2
    assert lc_data.lc_data_t0 == 3.0
    assert lc_data.filters == ["u", "g"]
    assert np.allclose(lc_data.lightcurves["u"][:, 0], [2.0, 3.0, 4.0, 5.0, 6.0])
    assert np.allclose(lc_data.lightcurves["g"][:, 0], [2.0, 3.0, 4.0, 5.0, 6.0])


def test_create_lightcurve_band_data_from_lclib_table_periodic() -> None:
    """Test that we can create a periodic LightcurveBandData object from a LCLIB table."""
    # When creating the LightcurveBandData from a table, we specifc the values in magnitude,
    # instead of flux.
    data = {
        "time": [0.0, 1.0, 2.0, 3.0, 4.0],
        "type": ["S", "S", "S", "S", "S"],
        "u": [1.0, 2.0, 1.0, 1.5, 1.0],
        "g": [2.0, 3.0, 2.0, 2.5, 2.0],
        "i": [1.0, 2.0, 3.0, 4.0, 5.0],  # First does not match last.
    }
    table = Table(data)
    table.meta["RECUR_CLASS"] = "RECUR-PERIODIC"
    lc_data = LightcurveBandData.from_lclib_table(table)

    # Check the internal structure of the LightcurveBandData.
    assert len(lc_data) == 3
    assert lc_data.lc_data_t0 == 0.0
    assert np.allclose(lc_data.period, 5.0)
    assert lc_data.filters == ["u", "g", "i"]

    # The first 5 entries match the entry data and the last was inserted
    # to match the periodicity.
    for filt in ["u", "g", "i"]:
        assert len(lc_data.lightcurves[filt]) == 6
        assert np.allclose(lc_data.lightcurves[filt][:5, 0], data["time"])
        assert np.allclose(lc_data.lightcurves[filt][5, 0], 5.0)

        assert np.allclose(
            lc_data.lightcurves[filt][:5, 1],
            mag2flux(np.array(data[filt])),
        )
        assert np.allclose(
            lc_data.lightcurves[filt][5, 1],
            lc_data.lightcurves[filt][0, 1],
        )


def test_create_lightcurve_template_model() -> None:
    """Test that we can create a simple LightcurveTemplateModel object."""
    pb_group = _create_toy_passbands()
    lightcurves = _create_toy_lightcurves()
    lc_model = LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=0.0, t0=0.0)
    assert lc_model.minphase(filter="u") == 0.8
    assert lc_model.minphase(filter="g") == 0.9
    assert lc_model.minphase(filter="r") == 1.0
    assert lc_model.maxphase(filter="u") == 10.8
    assert lc_model.maxphase(filter="g") == 10.9
    assert lc_model.maxphase(filter="r") == 11.0
    with pytest.raises(ValueError):
        lc_model.minphase()  # No filter provided
    with pytest.raises(ValueError):
        lc_model.maxphase()  # No filter provided

    # Check the internal structure of the LightcurveTemplateModel.
    assert len(lc_model.lightcurves) == 3
    assert lc_model.filters == ["u", "g", "r"]

    # A call to evaluate_bandfluxes should return the desired light curves.  We only use two of the passbands.
    graph_state = lc_model.sample_parameters(num_samples=1)
    query_times = np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.0, 20.0, 21.0])
    query_filters = np.array(["u", "r", "u", "r", "u", "r", "u", "r"])

    # Catch the warning since there are times outside the range and no baseline is provided.
    with pytest.warns(UserWarning):
        fluxes = lc_model.evaluate_bandfluxes(pb_group, query_times, query_filters, graph_state)
    assert len(fluxes) == len(query_times)

    # Timesteps 0.0, 0.5, 20.0, and 21.0 fall outside the range of the model and are set to 0.0.
    # Timesteps 1.0 and 3.0 are from the u band which is constant at 2.
    # Timesteps 2.0 and 4.0 are from the r band which is linearly increasing with time.
    assert np.allclose(fluxes, [0.0, 0.0, 2.0, 1.2, 2.0, 1.4, 0.0, 0.0])

    # A call to evaluate_bandfluxes with an unsupported filter should raise an error.
    query_filters = np.array(["u", "r", "u", "r", "u", "x", "u", "r", "r"])
    with pytest.raises(ValueError):
        lc_model.evaluate_bandfluxes(pb_group, query_times, query_filters, graph_state)


def test_create_lightcurve_template_model_no_pb() -> None:
    """Test that we can create a simple LightcurveTemplateModel object without passbands."""
    lightcurves = _create_toy_lightcurves()
    lc_model = LightcurveTemplateModel(lightcurves, None, lc_data_t0=0.0, t0=0.0)

    # Check the internal structure of the LightcurveTemplateModel.
    assert len(lc_model.lightcurves) == 3
    assert lc_model.sed_basis is None
    assert lc_model.filters == ["u", "g", "r"]

    # We can still evaluate the bandfluxes.
    graph_state = lc_model.sample_parameters(num_samples=1)
    query_times = np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.0, 20.0, 21.0])
    query_filters = np.array(["u", "r", "u", "r", "u", "r", "u", "r"])
    pb_group = _create_toy_passbands()

    with pytest.warns(UserWarning):
        fluxes = lc_model.evaluate_bandfluxes(pb_group, query_times, query_filters, graph_state)
    assert np.allclose(fluxes, [0.0, 0.0, 2.0, 1.2, 2.0, 1.4, 0.0, 0.0])

    # We fail if we try to compute and SED or plot the basis functions.
    with pytest.raises(ValueError):
        lc_model.compute_sed(query_times, np.array([5000.0, 6000.0]), graph_state)
    with pytest.raises(ValueError):
        lc_model.plot_sed_basis()


def test_create_lightcurve_template_model_from_data() -> None:
    """Test that we can create a simple LightcurveTemplateModel object for LightcurveBandData."""
    pb_group = _create_toy_passbands()
    lightcurves = _create_toy_lightcurves()
    lc_data = LightcurveBandData(lightcurves, lc_data_t0=0.0)
    lc_model = LightcurveTemplateModel(lc_data, pb_group, lc_data_t0=0.0, t0=0.0)

    # Check the internal structure of the LightcurveTemplateModel.
    assert len(lc_model.lightcurves) == 3
    assert lc_model.filters == ["u", "g", "r"]

    # A call to evaluate_bandfluxes should return the desired light curves.  We only use two of the passbands.
    graph_state = lc_model.sample_parameters(num_samples=1)
    query_times = np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.0, 20.0, 21.0])
    query_filters = np.array(["u", "r", "u", "r", "u", "r", "u", "r"])

    with pytest.warns(UserWarning):
        fluxes = lc_model.evaluate_bandfluxes(pb_group, query_times, query_filters, graph_state)
    assert len(fluxes) == len(query_times)

    # Timesteps 0.0, 0.5, 20.0, and 21.0 fall outside the range of the model and are set to 0.0.
    # Timesteps 1.0 and 3.0 are from the u band which is constant at 2.
    # Timesteps 2.0 and 4.0 are from the r band which is linearly increasing with time.
    assert np.allclose(fluxes, [0.0, 0.0, 2.0, 1.2, 2.0, 1.4, 0.0, 0.0])

    # A call to evaluate_bandfluxes with an unsupported filter should raise an error.
    query_filters = np.array(["u", "r", "u", "r", "u", "x", "u", "r", "r"])
    with pytest.raises(ValueError):
        lc_model.evaluate_bandfluxes(pb_group, query_times, query_filters, graph_state)


def test_create_lightcurve_template_model_unsorted() -> None:
    """Test that we fail if we try to create a LightcurveTemplateModel with unsorted light curves."""
    pb_group = _create_toy_passbands()
    lightcurves = _create_toy_lightcurves()
    lightcurves["u"][0, 0] = 2.0  # Make the u band unsorted

    with pytest.raises(ValueError):
        # We should fail because the light curves are not sorted by time.
        LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=0.0, t0=0.0)


def test_create_lightcurve_template_model_baseline() -> None:
    """Test that we can create a simple LightcurveTemplateModel object with baseline values."""
    pb_group = _create_toy_passbands()
    lightcurves = _create_toy_lightcurves()
    baseline = {"u": 0.5, "g": 1.2, "r": 0.05}
    lc_model = LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=0.0, t0=0.0, baseline=baseline)

    # A call to evaluate_bandfluxes should return the desired light curves.
    # We only use two of the passbands.
    graph_state = lc_model.sample_parameters(num_samples=1)
    query_times = np.array([-100.0, 0.0, 0.5, 1.0, 2.0, 3.0, 4.0, 20.0, 21.0, 1000.0])
    query_filters = np.array(["u", "u", "r", "u", "r", "u", "r", "u", "r", "r"])
    fluxes = lc_model.evaluate_bandfluxes(pb_group, query_times, query_filters, graph_state)
    assert len(fluxes) == len(query_times)

    # Timesteps -100.0, 0.0, 0.5, 20.0, 21.0, 1000.0 all fall outside the range of the model
    # and use baseline values for the respective bands.
    # Timesteps 1.0 and 3.0 are from the u band which is constant at 2.
    # Timesteps 2.0 and 4.0 are from the r band which is linearly increasing with time.
    assert np.allclose(fluxes, [0.5, 0.5, 0.05, 2.0, 1.2, 2.0, 1.4, 0.5, 0.05, 0.05])

    # We fail if we try to create a LightcurveTemplateModel with a baseline that does
    # not match the passbands (no r band provided).
    with pytest.raises(ValueError):
        LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=0.0, t0=0.0, baseline={"u": 0.5, "g": 1.2})

    # Test that we can add a constant dimming effect and it is applied in to the bandpass values.
    effect = ScaleFluxEffect(flux_scale=0.1)
    lc_model.add_effect(effect)
    graph_state = lc_model.sample_parameters(num_samples=1)

    fluxes = lc_model.evaluate_bandfluxes(pb_group, query_times, query_filters, graph_state)
    assert len(fluxes) == len(query_times)
    assert np.allclose(fluxes, [0.05, 0.05, 0.005, 0.20, 0.12, 0.20, 0.14, 0.05, 0.005, 0.005])


def test_create_lightcurve_template_model_periodic() -> None:
    """Test that we can create a periodic LightcurveTemplateModel object."""
    pb_group = _create_toy_passbands()

    with pytest.raises(ValueError):
        # We cannot create a periodic model from light curves that do
        # not cover the same time range.
        lightcurves = _create_toy_lightcurves()
        LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=0.0, periodic=True)

    times = np.arange(0.0, 10.5, 0.5)
    g_curve = 3.0 * np.ones_like(times)
    r_curve = 0.1 * times + 1.0
    r_curve[-1] = r_curve[0]  # Make sure the 1st and last values are the same
    lightcurves = {
        "g": np.array([times, g_curve]).T,
        "r": np.array([times, r_curve]).T,
    }
    lc_model = LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=0.0, t0=0.0, periodic=True)

    # A call to evaluate_bandfluxes should return the desired light curves.
    graph_state = lc_model.sample_parameters(num_samples=1)
    query_times = np.array([1.0, 5.0, 11.0, 15.0, 21.0, 25.0, 51.0])
    query_filters = np.full(len(query_times), "r")
    fluxes = lc_model.evaluate_bandfluxes(pb_group, query_times, query_filters, graph_state)
    assert len(fluxes) == len(query_times)

    # Time steps 1.0, 11.0, 21.0, and 51.0 are all wrapped to the same point.
    # Time steps 5.0, 15.0, and 25.0 are all wrapped to the same point.
    assert np.allclose(fluxes, [1.1, 1.5, 1.1, 1.5, 1.1, 1.5, 1.1])

    # Check a curve if defined with a first time > 0.0.
    times = np.array([2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0])
    g_curve = 3.0 * np.ones_like(times)
    r_curve = [0.0, 1.0, 2.0, 3.0, 2.5, 1.5, 0.0]
    lightcurves = {
        "g": np.array([times, g_curve]).T,
        "r": np.array([times, r_curve]).T,
    }

    # We fail if we specify an incorrect lc_data_t0 for a periodic light curve.
    with pytest.raises(ValueError):
        _ = LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=1.0, t0=0.0, periodic=True)

    lc_model = LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=2.0, t0=0.0, periodic=True)
    query_times = np.array([0.0, 1.0, 2.0, 3.0, 4.0])
    query_filters = np.full(len(query_times), "r")

    # A call to evaluate_bandfluxes should return the desired light curves. Since t0=0.0, query
    # time 0.0 should correspond to the start of the light curve's period.
    graph_state = lc_model.sample_parameters(num_samples=1)
    fluxes = lc_model.evaluate_bandfluxes(pb_group, query_times, query_filters, graph_state)
    assert np.allclose(fluxes, [0.0, 1.0, 2.0, 3.0, 2.5])

    # We can also auto-derive lc_data_t0 from the light curves.
    lc_model = LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=0.0, t0=0.0, periodic=True)
    graph_state = lc_model.sample_parameters(num_samples=1)
    fluxes = lc_model.evaluate_bandfluxes(pb_group, query_times, query_filters, graph_state)
    assert np.allclose(fluxes, [0.0, 1.0, 2.0, 3.0, 2.5])

    # If we use t0=1.0, we are saying the period starts at 1.0 for this sample, so a
    # query time of 0.0 should wrap around and return the *last* value of the light curve.
    lc_model = LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=0.0, t0=1.0, periodic=True)
    graph_state = lc_model.sample_parameters(num_samples=1)
    fluxes = lc_model.evaluate_bandfluxes(pb_group, query_times, query_filters, graph_state)
    assert np.allclose(fluxes, [1.5, 0.0, 1.0, 2.0, 3.0])


def test_create_lightcurve_template_model_periodic_complex_offsets() -> None:
    """Test that we can create a periodic LightcurveTemplateModel with complex offsets."""
    pb_group = _create_toy_passbands()

    # Create a light curve with 20 samples over a 10 day time range.
    dt = np.arange(0.0, 10.5, 0.5)
    r_curve = np.abs(dt - 5.0)  # Sawtooth-like curve starting and ending at 5.0
    g_curve = 3.0 * np.ones_like(dt)

    # Define the actual times for the light curves as starting at 60676.0.
    times = 60676.0 + dt
    lightcurves = {
        "g": np.array([times, g_curve]).T,
        "r": np.array([times, r_curve]).T,
    }

    # Create a LightcurveTemplateModel with t0=60672.0, so we are shifting it back by 4 days.
    lc_model = LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=0.0, t0=60672.0, periodic=True)
    graph_state = lc_model.sample_parameters(num_samples=1)

    # Check query times relative to 60676.0 (4 days after the period started).
    query_times = 60676.0 + np.array([0.0, 0.5, 1.0, 2.0, 6.5, 12.0])
    query_filters = np.full(len(query_times), "r")
    fluxes = lc_model.evaluate_bandfluxes(pb_group, query_times, query_filters, graph_state)
    assert np.allclose(fluxes, [1.0, 0.5, 0.0, 1.0, 4.5, 1.0])


def test_create_lightcurve_template_model_numpy() -> None:
    """Test that we can create a simple LightcurveTemplateModel object from a numpy array."""
    pb_group = _create_toy_passbands()

    # Create fake light curves over the time range 0.0 to 4.3. The u band is linearly
    # decreasing, the g band is constant, and the r band is linearly increasing.
    lightcurves = np.array(
        [
            [0.0, 10.0, "u"],
            [0.1, 11.0, "g"],
            [0.3, 11.0, "r"],
            [1.0, 10.1, "u"],
            [1.1, 11.0, "g"],
            [1.3, 10.9, "r"],
            [2.0, 10.2, "u"],
            [2.1, 11.0, "g"],
            [2.3, 10.8, "r"],
            [3.0, 10.3, "u"],
            [3.1, 11.0, "g"],
            [3.3, 10.7, "r"],
            [4.0, 10.4, "u"],
            [4.1, 11.0, "g"],
            [4.3, 10.6, "r"],
        ]
    )
    lc_model = LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=0.0, t0=0.0)

    # Check the internal structure of the LightcurveTemplateModel.
    assert len(lc_model.lightcurves) == 3
    assert set(lc_model.filters) == set(["u", "g", "r"])

    # A call to evaluate_bandfluxes should return the desired light curves.
    graph_state = lc_model.sample_parameters(num_samples=1)
    query_times = np.array([0.5, 0.6, 1.8, 2.3, 2.8, 3.0, 3.5, 4.0])
    query_filters = np.array(["u", "g", "r", "r", "r", "g", "u", "u"])
    expected = np.array([10.05, 11.0, 10.85, 10.8, 10.75, 11.0, 10.35, 10.4])

    fluxes = lc_model.evaluate_bandfluxes(pb_group, query_times, query_filters, graph_state)
    assert len(fluxes) == len(expected)
    assert np.allclose(fluxes, expected)


def test_create_lightcurve_template_model_t0() -> None:
    """Test that we can create a simple LightcurveTemplateModel object with a given t0."""
    pb_group = _create_toy_passbands()
    lightcurves = _create_toy_lightcurves()
    lc_model = LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=0.0, t0=60676.0)

    graph_state = lc_model.sample_parameters(num_samples=1)  # needed for t0
    query_times = 60676.0 + np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.0, 20.0, 21.0])
    query_filters = np.array(["u", "r", "u", "r", "u", "r", "u", "r"])

    with pytest.warns(UserWarning):
        fluxes = lc_model.evaluate_bandfluxes(pb_group, query_times, query_filters, graph_state)
    assert len(fluxes) == len(query_times)

    # Timesteps +0.0, +0.5, +20.0, and +21.0 fall outside the range of the model and are set to 0.0.
    # Timesteps +1.0 and +3.0 are from the u band which is constant at 2.
    # Timesteps +2.0 and +4.0 are from the r band which is linearly increasing with time
    # as 0.1 * t + 1.0.
    assert np.allclose(fluxes, [0.0, 0.0, 2.0, 1.2, 2.0, 1.4, 0.0, 0.0])

    # Test that we can also handle a light curve with a different lc_data_t0.
    lc_model2 = LightcurveTemplateModel(lightcurves, pb_group, t0=60676.0, lc_data_t0=1.0)
    graph_state2 = lc_model2.sample_parameters(num_samples=1)  # needed for t0
    with pytest.warns(UserWarning):
        fluxes2 = lc_model2.evaluate_bandfluxes(pb_group, query_times, query_filters, graph_state2)

    # Timesteps +20.0 and +21.0 fall outside the range of the model and are set to 0.0.
    # Timesteps +0.0, +1.0 and +3.0 correspond to the original times (in the light curve definition)
    # of +1.0, +2.0, and +3.0 which are from the u band which is constant at 2.
    # Timesteps +0.5, +2.0 and +4.0 correspond to the original times (in the light curve definition)
    # of +1.5, +2.0, and +5.0 which are from the r band which is linearly increasing with time
    # as 0.1 * t + 1.0.
    assert np.allclose(fluxes2, [2.0, 1.15, 2.0, 1.3, 2.0, 1.5, 0.0, 0.0])


def test_create_lightcurve_template_model_extrapolation() -> None:
    """Test that we can create a simple LightcurveTemplateModel object with a given t0
    and time extrapolation outside those bounds."""
    # We offset "u" by -0.5 days, so it uses a different ending point value
    # for its time extrapolation than the "r" band.
    times = np.linspace(1, 11, 20)
    lightcurves = {
        "u": np.array([times - 0.5, 2.0 * np.ones_like(times)]).T,
        "g": np.array([times, 3.0 * np.ones_like(times)]).T,
        "r": np.array([times, 0.1 * times + 1.0]).T,
    }
    extrap = LinearDecay(decay_width=10.0)  # Decay to 0 over 10 days
    lc_model = LightcurveTemplateModel(
        lightcurves,
        None,  # No passband data
        lc_data_t0=0.0,
        t0=60676.0,
        time_extrapolation=extrap,
    )

    graph_state = lc_model.sample_parameters(num_samples=1)  # needed for t0
    query_times = 60676.0 + np.array([-1.0, -0.5, 0.0, 0.5, 1.0, 2.0, 3.0, 4.0, 20.0, 21.0])
    query_filters = np.array(["u", "r", "u", "r", "u", "r", "u", "r", "u", "r"])

    # No warning because we have extrapolation.
    fluxes = lc_model.evaluate_bandfluxes(None, query_times, query_filters, graph_state)
    assert len(fluxes) == len(query_times)

    # Timesteps +1.0 and +3.0 are from the u band which is constant at 2.
    # Timesteps +2.0 and +4.0 are from the r band which is linearly increasing with time
    # as 0.1 * t + 1.0.
    # Timesteps -1.0, -0.5, +0.0, +0.5, +20.0, and +21.0 fall outside the range of the model
    # and use linear extrapolation from their respective edge values in the bands.
    assert np.allclose(fluxes, [1.7, 0.935, 1.9, 1.045, 2.0, 1.2, 2.0, 1.4, 0.1, 0.0])

    lc_model2 = LightcurveTemplateModel(
        lightcurves,
        None,  # No passband data
        lc_data_t0=0.0,
        t0=60676.0,
        time_extrapolation=(extrap, None),  # Only extrapolate on the left
    )
    graph_state2 = lc_model2.sample_parameters(num_samples=1)  # needed for t0
    with pytest.warns(UserWarning):
        fluxes2 = lc_model2.evaluate_bandfluxes(None, query_times, query_filters, graph_state2)
    assert np.allclose(fluxes2, [1.7, 0.935, 1.9, 1.045, 2.0, 1.2, 2.0, 1.4, 0.0, 0.0])

    lc_model3 = LightcurveTemplateModel(
        lightcurves,
        None,  # No passband data
        lc_data_t0=0.0,
        t0=60676.0,
        time_extrapolation=(None, extrap),  # Only extrapolate on the right
    )
    graph_state3 = lc_model3.sample_parameters(num_samples=1)  # needed for t0
    with pytest.warns(UserWarning):
        fluxes3 = lc_model3.evaluate_bandfluxes(None, query_times, query_filters, graph_state3)
    assert np.allclose(fluxes3, [0.0, 0.0, 0.0, 0.0, 2.0, 1.2, 2.0, 1.4, 0.1, 0.0])


def test_create_lightcurve_template_model_fail() -> None:
    """Test fail cases for creating the LightcurveTemplateModel object."""
    a_band = Passband(np.array([[400, 0.5], [500, 0.5], [600, 0.5]]), "LSST", "u")
    b_band = Passband(np.array([[800, 0.8], [900, 0.8], [1000, 0.8]]), "LSST", "g")
    c_band = Passband(np.array([[900, 0.6], [1000, 0.6], [1100, 0.6]]), "LSST", "r")
    pb_group = PassbandGroup(given_passbands=[a_band, b_band, c_band])

    times = np.linspace(0, 10, 100)
    lightcurves = {
        "u": np.array([times, 2.0 * np.ones_like(times)]).T,
        "g": np.array([times, 3.0 * np.ones_like(times)]).T,
        "r": np.array([times, 4.0 * np.ones_like(times)]).T,
        "i": np.array([times, 0.1 * times + 1.0]).T,
    }

    # Fail on mismatched passbands.
    with pytest.raises(ValueError):
        _ = LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=0.0, t0=0.0)

    # Remove the offending passband and try again.
    del lightcurves["i"]
    _ = LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=0.0, t0=0.0)

    # We fail without a t0 value.
    with pytest.raises(ValueError):
        _ = LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=0.0)

    # Make one of the lightcurves the wrong shape.
    lightcurves["u"] = times.T
    with pytest.raises(ValueError):
        _ = LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=0.0, t0=0.0)
    lightcurves["u"] = np.array([times, 2.0 * np.ones_like(times)]).T

    # We fail if two passbands overlap other passbands completely.
    d_band = Passband(np.array([[850, 0.6], [1050, 0.6]]), "LSST", "i")
    pb_group = PassbandGroup(given_passbands=[a_band, b_band, c_band, d_band])
    lightcurves["i"] = np.array([times, 0.1 * times + 1.0]).T
    with pytest.raises(ValueError):
        _ = LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=0.0, t0=0.0)


def test_lightcurve_plot() -> None:
    """Test that the plotting functions do not crash."""
    pb_group = _create_toy_passbands()
    lightcurves = _create_toy_lightcurves()
    lc_model = LightcurveTemplateModel(lightcurves, pb_group, lc_data_t0=0.0, t0=0.0)
    lc_model.plot_lightcurves()
    lc_model.plot_sed_basis()


def test_create_multilightcurve_template_model() -> None:
    """Test that we can create a simple MultiLightcurveTemplateModel object."""
    pb_group = _create_toy_passbands()

    # Lightcurve 1 is non-periodic and covers u and g.
    lc1_times = np.arange(0.0, 10.5, 0.5)
    lc1_lightcurves = {
        "u": np.array([lc1_times + 0.1, 2.0 * np.ones_like(lc1_times)]).T,
        "g": np.array([lc1_times, 3.0 * np.ones_like(lc1_times)]).T,
    }
    lc1_data = LightcurveBandData(lc1_lightcurves, lc_data_t0=0.0, baseline={"u": 0.1, "g": 0.2})

    # Light curve 2 is periodic and covers r and g.
    lc2_times = np.arange(0.0, 19.0, 1.0)
    lc2_lightcurves = {
        "r": np.array([lc2_times, lc2_times % 2]).T,
        "g": np.array([lc2_times, lc2_times % 2 + 0.5]).T,
    }
    lc2_data = LightcurveBandData(lc2_lightcurves, lc_data_t0=0.0, periodic=True)

    # Create the MultiLightcurveTemplateModel with both light curves.
    model = MultiLightcurveTemplateModel(
        [lc1_data, lc2_data],
        pb_group,
        weights=[0.25, 0.75],
        t0=0.0,
        node_label="source",
    )
    assert len(model.lightcurves) == 2
    assert set(model.filters) == {"u", "g", "r"}

    # Check that we sample the light curves correctly.
    graph_state = model.sample_parameters(num_samples=1_000)
    lc_used = graph_state["source"]["selected_lightcurve"]
    assert np.all((lc_used == 0) | (lc_used == 1))
    assert np.sum(lc_used == 0) >= 200  # approximately 25% of the samples should be from lc1
    assert np.sum(lc_used == 1) >= 700  # approximately 75% of the samples should be from lc2

    # Check that the baseline values are set correctly for the selected light curves.
    # Light curve 0 has values given for u and g. Light curve 1 does not have any given
    # baseline values, so it should default to 0.0 in all bands.
    lc0_mask = lc_used == 0
    assert np.all(graph_state["source"]["baseline_u"][lc0_mask] == 0.1)
    assert np.all(graph_state["source"]["baseline_g"][lc0_mask] == 0.2)
    assert np.all(graph_state["source"]["baseline_r"][lc0_mask] == 0.0)  # Default

    lc1_mask = lc_used == 1
    assert np.all(graph_state["source"]["baseline_u"][lc1_mask] == 0.0)
    assert np.all(graph_state["source"]["baseline_g"][lc1_mask] == 0.0)
    assert np.all(graph_state["source"]["baseline_r"][lc1_mask] == 0.0)

    query_times = np.array([-1.0, 1.0, 2.0, 3.0, 4.0, 20.0, 21.0])
    query_filters = np.full(len(query_times), "g")
    fluxes = model.evaluate_bandfluxes(pb_group, query_times, query_filters, graph_state)
    assert len(fluxes) == 1_000

    # Check that we sampled from the light curve that we said we did.
    expected_0 = np.array([0.2, 3.0, 3.0, 3.0, 3.0, 0.2, 0.2])
    expected_1 = np.array([1.5, 1.5, 0.5, 1.5, 0.5, 0.5, 1.5])
    for idx in range(1_000):
        if lc_used[idx] == 0:
            # The first light curve is 3.0 when active
            assert np.allclose(fluxes[idx], expected_0)
        else:
            assert np.allclose(fluxes[idx], expected_1)

    # We cannot compute phase bounds from a graph state with 1000 samples.
    with pytest.raises(ValueError):
        _ = model.minphase(filter="g", graph_state=graph_state)
    with pytest.raises(ValueError):
        _ = model.maxphase(filter="g", graph_state=graph_state)

    # A call to evaluate_bandfluxes with an unsupported filter should raise an error.
    query_filters[2] = "x"  # Change one filter to an unsupported one
    with pytest.raises(ValueError):
        _ = model.evaluate_bandfluxes(pb_group, query_times, query_filters, graph_state)

    # Check that we can get the bounds for a single light curve. Neither has bounds because
    # model 0 has a baseline and model 1 is periodic.
    single_state = model.sample_parameters(num_samples=1)
    assert model.minphase(filter="g", graph_state=single_state) is None
    assert model.maxphase(filter="g", graph_state=single_state) is None


def test_create_multilightcurve_template_model_fail() -> None:
    """Test creating a MultiLightcurveTemplateModel with invalid parameters."""
    pb_group = _create_toy_passbands()

    # Light curve 1 is non-periodic and covers u and g.
    lc1_times = np.arange(0.0, 10.5, 0.5)
    lc1_lightcurves = {
        "u": np.array([lc1_times + 0.1, 2.0 * np.ones_like(lc1_times)]).T,
        "g": np.array([lc1_times, 3.0 * np.ones_like(lc1_times)]).T,
        "r": np.array([lc1_times, 3.0 * np.ones_like(lc1_times)]).T,
    }
    lc1_data = LightcurveBandData(lc1_lightcurves, lc_data_t0=0.0, baseline={"u": 0.1, "g": 0.2, "r": 0.3})

    # This single template works.
    _ = MultiLightcurveTemplateModel([lc1_data], pb_group, t0=0.0)

    # We fail with no t0 value.
    with pytest.raises(ValueError):
        _ = MultiLightcurveTemplateModel([lc1_data], pb_group)

    # Light curve 2 is periodic and covers r and g.
    lc2_times = np.arange(0.0, 19.0, 1.0)
    lc2_lightcurves = {
        "r": np.array([lc2_times, lc2_times % 2]).T,
        "g": np.array([lc2_times, lc2_times % 2 + 0.5]).T,
        "i": np.array([lc2_times, lc2_times % 2 + 0.5]).T,
    }
    lc2_data = LightcurveBandData(lc2_lightcurves, lc_data_t0=0.0, periodic=True)

    with pytest.raises(ValueError):
        # The passband group does not have data for the 'i' band.
        _ = MultiLightcurveTemplateModel([lc1_data, lc2_data], pb_group, weights=[0.25, 0.75], t0=0.0)


def test_create_multilightcurve_from_lclib_file(test_data_dir):
    """Test creating a MultiLightcurveTemplateModel from a LCLIB file."""
    passband_list = []
    pb_start = np.array([[400, 0.5], [500, 0.5], [600, 0.5]])
    pb_shift = np.array([[500, 0], [500, 0], [500, 0]])
    for idx, filter in enumerate(["u", "g", "r", "i", "z"]):
        pb = Passband(pb_start + idx * pb_shift, "LSST", filter)
        passband_list.append(pb)
    pb_group = PassbandGroup(given_passbands=passband_list)

    lc_file = test_data_dir / "test_lclib_data.TEXT"
    model = MultiLightcurveTemplateModel.from_lclib_file(lc_file, pb_group, t0=0.0)
    assert len(model.lightcurves) == 3
    assert set(model.filters) == {"u", "g", "r", "i", "z"}
    for lc in model.lightcurves:
        assert lc.lc_data_t0 > 0.0

    # Confirm that we have noted the data in the citations registry.
    citations = find_in_citations("LCLIB Data")
    assert len(citations) == 1
    assert str(lc_file) in citations[0]

    # Check that we can override the times.
    forced_lc_t0 = np.array([0.0, 1.0, 2.0])
    model = MultiLightcurveTemplateModel.from_lclib_file(lc_file, pb_group, t0=0.0, forced_lc_t0=forced_lc_t0)
    assert len(model.lightcurves) == 3
    for idx, lc in enumerate(model.lightcurves):
        assert lc.lc_data_t0 == forced_lc_t0[idx]

    # Check that we can override the times with a scalar.
    model = MultiLightcurveTemplateModel.from_lclib_file(lc_file, pb_group, t0=0.0, forced_lc_t0=5.0)
    assert len(model.lightcurves) == 3
    for lc in model.lightcurves:
        assert lc.lc_data_t0 == 5.0


def test_create_multilightcurve_from_lclib_file_time(test_data_dir):
    """Test creating a MultiLightcurveTemplateModel from a LCLIB file."""
    passband_list = []
    pb_start = np.array([[400, 0.5], [500, 0.5], [600, 0.5]])
    pb_shift = np.array([[500, 0], [500, 0], [500, 0]])
    for idx, filter in enumerate(["u", "g", "r", "i", "z"]):
        pb = Passband(pb_start + idx * pb_shift, "LSST", filter)
        passband_list.append(pb)
    pb_group = PassbandGroup(given_passbands=passband_list)

    lc_file = test_data_dir / "test_lclib_data.TEXT"
    model = MultiLightcurveTemplateModel.from_lclib_file(lc_file, pb_group, t0=0.0)
    assert len(model.lightcurves) == 3
    assert set(model.filters) == {"u", "g", "r", "i", "z"}


def test_create_multilightcurve_from_lclib_file_filtered(test_data_dir):
    """Test creating a MultiLightcurveTemplateModel from a LCLIB file using a subset of filters."""
    passband_list = []
    pb_start = np.array([[400, 0.5], [500, 0.5], [600, 0.5]])
    pb_shift = np.array([[500, 0], [500, 0], [500, 0]])
    for idx, filter in enumerate(["u", "g", "r", "i", "z"]):
        pb = Passband(pb_start + idx * pb_shift, "LSST", filter)
        passband_list.append(pb)
    pb_group = PassbandGroup(given_passbands=passband_list)

    lc_file = test_data_dir / "test_lclib_data.TEXT"
    model = MultiLightcurveTemplateModel.from_lclib_file(lc_file, pb_group, t0=0.0, filters=["u", "g", "r"])
    assert len(model.lightcurves) == 3
    assert set(model.filters) == {"u", "g", "r"}



import argparse

import numpy as np
from lightcurvelynx.obstable.opsim import OpSim

_all_columns = [
    "observationId",
    "fieldRA",
    "fieldDec",
    "observationStartMJD",
    "visitExposureTime",
    "filter",
    "rotSkyPos",
    "numExposures",
    "airmass",
    "seeingFwhmEff",
    "seeingFwhmGeom",
    "skyBrightness",
]


def make_sampled_opsim(ra_vals, dec_vals, time_step, num_visits):
    """Create a fake OpSim data that scans a grid of (ra, dec) multiple times.

    Note
    ----
    This is a toy OpSim with lots of simplifying assumptions. It should not be
    used for scientific validation.

    Parameters
    ----------
    ra_vals : `numpy.ndarray`
       The values of RA to sample (in degrees).
    dec_vals : `numpy.ndarray`
       The values of dec to sample (in degrees).
    time_step : `float`
        The time between observations (in days).
    num_visits : `int`
        The number of times the full grid is sampled.
    """
    opsim_data = {
        "observationStartMJD": [],
        "fieldRA": [],
        "fieldDec": [],
    }

    for day in range(num_visits):
        t = 60676.0 + 5.0 * day  # Start date = Jan 1, 2025
        for ra in ra_vals:
            for dec in dec_vals:
                opsim_data["observationStartMJD"].append(t)
                opsim_data["fieldRA"].append(ra)
                opsim_data["fieldDec"].append(dec)
                t += time_step

    # Add other fields
    num_obs = len(opsim_data["observationStartMJD"])
    opsim_data["observationId"] = np.arange(num_obs)
    opsim_data["filter"] = np.repeat(["g", "r"], num_obs // 2)
    opsim_data["visitExposureTime"] = np.full((num_obs), 29.2)  # Most common value
    opsim_data["numExposures"] = np.full((num_obs), 2)  # Most common value
    opsim_data["airmass"] = np.full((num_obs), 1.3)  # Mean value
    opsim_data["seeingFwhmEff"] = np.full((num_obs), 1.12)  # Mean value
    opsim_data["seeingFwhmGeom"] = np.full((num_obs), 0.97)  # Mean value
    opsim_data["skyBrightness"] = np.full((num_obs), 20.0)  # Mean value

    for col in _all_columns:
        if col not in opsim_data:
            opsim_data[col] = np.zeros(num_obs)

    return OpSim(opsim_data)


def main():
    """Generate the fake OpSim data and save it to a file.

    To generate an updated small_db file use:
        python tests/lightcurvelynx/utils/make_fake_opsim.py tests/lightcurvelynx/data/opsim_small.db
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("output", type=str, help="Output filename")
    parser.add_argument(
        "--ra_start", type=float, default=0.0, help="The starting RA of the grid (in degrees)"
    )
    parser.add_argument("--ra_end", type=float, default=30.0, help="The end RA of the grid (in degrees)")
    parser.add_argument("--ra_step", type=float, default=3.0, help="The grid step size along RA (in degrees)")
    parser.add_argument(
        "--dec_start", type=float, default=-30.0, help="The starting dec of the grid (in degrees)"
    )
    parser.add_argument("--dec_end", type=float, default=0.0, help="The end dec of the grid (in degrees)")
    parser.add_argument(
        "--dec_step", type=float, default=3.0, help="The grid step size along dec (in degrees)"
    )
    parser.add_argument(
        "--time_step", type=float, default=0.003, help="The time between observations (in days)"
    )
    parser.add_argument(
        "--num_visits", type=int, default=3, help="The number of times the full grid is sampled"
    )
    args = parser.parse_args()

    # Create the sample points for RA and dec.
    ra_vals = np.arange(args.ra_start, args.ra_end, args.ra_step)
    dec_vals = np.arange(args.dec_start, args.dec_end, args.dec_step)
    print(f"Sampling along\n  ra={ra_vals}\n  dec={dec_vals}")

    # Create the OpSim and save it to a file.
    ops_data = make_sampled_opsim(ra_vals, dec_vals, args.time_step, args.num_visits)
    ops_data.write_opsim_table(args.output)


if __name__ == "__main__":
    main()



import numpy as np
from lightcurvelynx.example_runs.simulate_snia import run_snia_end2end


def test_snia_end2end(oversampled_observations, passbands_dir):
    """Test that the end to end run works."""
    rng = np.random.default_rng(123)

    solid_angle = 0.0001
    res_list, passbands = run_snia_end2end(
        oversampled_observations,
        passbands_dir,
        solid_angle=solid_angle,
        nsample=4,
        rng_info=rng,
    )
    assert len(passbands) == 2
    assert len(res_list) == 4

    num_samples = 1
    res_list, passbands = run_snia_end2end(
        oversampled_observations,
        passbands_dir,
        solid_angle=solid_angle,
        nsample=num_samples,
        rng_info=rng,
    )
    assert len(res_list) == num_samples
    assert len(passbands) == 2
